from __future__ import annotations

import threading
from typing import TYPE_CHECKING, Any, NamedTuple

from psycopg.generators import notifies
from psycopg.sql import SQL, Composed, Identifier

from eventsourcing.dcb.api import (
    DCBAppendCondition,
    DCBEvent,
    DCBQuery,
    DCBQueryItem,
    DCBReadResponse,
    DCBRecorder,
    DCBSequencedEvent,
)
from eventsourcing.dcb.persistence import (
    DCBInfrastructureFactory,
    DCBListenNotifySubscription,
)
from eventsourcing.dcb.popo import SimpleDCBReadResponse
from eventsourcing.persistence import IntegrityError, InternalError, ProgrammingError
from eventsourcing.postgres import (
    NO_TRACEBACK,
    BasePostgresFactory,
    PostgresDatastore,
    PostgresRecorder,
    PostgresTrackingRecorder,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from psycopg import Connection, Cursor
    from psycopg.abc import Params
    from psycopg.rows import DictRow

DB_TYPE_NAME_DCB_EVENT_TT = "dcb_event_tt"

DB_TYPE_DCB_EVENT = SQL("""
CREATE TYPE {schema}.{event_type} AS (
    type text,
    data bytea,
    tags text[]
)
""")

DB_TYPE_NAME_DCB_QUERY_ITEM_TT = "dcb_query_item_tt"

DB_TYPE_DCB_QUERY_ITEM = SQL("""
CREATE TYPE {schema}.{query_item_type} AS (
    types text[],
    tags text[]
)
""")

DB_TABLE_DCB_EVENTS = SQL("""
CREATE TABLE IF NOT EXISTS {schema}.{events_table} (
    id bigserial,
    type text NOT NULL ,
    data bytea,
    tags text[] NOT NULL
) WITH (
  autovacuum_enabled = true,
  autovacuum_vacuum_threshold = 100000000,  -- Effectively disables VACUUM
  autovacuum_vacuum_scale_factor = 0.5,     -- Same here, high scale factor
  autovacuum_analyze_threshold = 1000,      -- Triggers ANALYZE more often
  autovacuum_analyze_scale_factor = 0.01    -- Triggers after 1% new rows
)
""")

DB_INDEX_UNIQUE_ID_COVER_TYPE = SQL("""
CREATE UNIQUE INDEX IF NOT EXISTS {id_cover_type_index} ON
{schema}.{events_table} (id) INCLUDE (type)
""")

DB_TABLE_DCB_TAGS = SQL("""
CREATE TABLE IF NOT EXISTS {schema}.{tags_table} (
    tag text,
    main_id bigint REFERENCES {events_table} (id)
) WITH (
    autovacuum_enabled = true,
    autovacuum_vacuum_threshold = 100000000,  -- Effectively disables VACUUM
    autovacuum_vacuum_scale_factor = 0.5,     -- Same here, high scale factor
    autovacuum_analyze_threshold = 1000,      -- Triggers ANALYZE more often
    autovacuum_analyze_scale_factor = 0.01    -- Triggers after 1% new rows
)
""")

DB_INDEX_TAG_MAIN_ID = SQL("""
CREATE INDEX IF NOT EXISTS {tag_main_id_index} ON
{schema}.{tags_table} (tag, main_id)
""")

SQL_SELECT_ALL = SQL("""
SELECT * FROM {schema}.{events_table}
WHERE id > COALESCE(%(after)s, 0)
ORDER BY id ASC
LIMIT COALESCE(%(limit)s, 9223372036854775807)
""")

SQL_SELECT_EVENTS_BY_TYPE = SQL("""
SELECT * FROM {schema}.{events_table}
WHERE type = %(event_type)s
AND id > COALESCE(%(after)s, 0)
ORDER BY id ASC
LIMIT COALESCE(%(limit)s, 9223372036854775807)
""")

SQL_SELECT_MAX_ID = SQL("""
SELECT MAX(id) FROM {schema}.{events_table}
""")

SQL_SELECT_BY_TAGS = SQL("""
WITH query_items AS (
    SELECT * FROM unnest(
        %(query_items)s::{schema}.{query_item_type}[]
    ) WITH ORDINALITY
),
initial_matches AS (
    SELECT
        t.main_id,
        qi.ordinality,
        t.tag,
        qi.tags AS required_tags,
        qi.types AS allowed_types
    FROM query_items qi
    JOIN {schema}.{tags_table} t
      ON t.tag = ANY(qi.tags)
   WHERE t.main_id > COALESCE(%(after)s, 0)
),
matched_groups AS (
    SELECT
        main_id,
        ordinality,
        COUNT(DISTINCT tag) AS matched_tag_count,
        array_length(required_tags, 1) AS required_tag_count,
        allowed_types
    FROM initial_matches
    GROUP BY main_id, ordinality, required_tag_count, allowed_types
),
qualified_ids AS (
    SELECT main_id, allowed_types
    FROM matched_groups
    WHERE matched_tag_count = required_tag_count
),
filtered_ids AS (
    SELECT m.id
    FROM {schema}.{events_table} m
    JOIN qualified_ids q ON q.main_id = m.id
    WHERE
        m.id > COALESCE(%(after)s, 0)
        AND (
            array_length(q.allowed_types, 1) IS NULL
            OR array_length(q.allowed_types, 1) = 0
            OR m.type = ANY(q.allowed_types)
        )
    ORDER BY m.id ASC
    LIMIT COALESCE(%(limit)s, 9223372036854775807)
)
SELECT *
FROM {schema}.{events_table}  m
WHERE m.id IN (SELECT id FROM filtered_ids)
ORDER BY m.id ASC;
""")


SQL_UNCONDITIONAL_APPEND = SQL("""
SELECT * FROM {schema}.{unconditional_append}(%(events)s)
""")
DB_FUNCTION_NAME_DCB_UNCONDITIONAL_APPEND_TT = "dcb_unconditional_append_tt"
DB_FUNCTION_UNCONDITIONAL_APPEND = SQL("""
CREATE OR REPLACE FUNCTION {schema}.{unconditional_append}(
    new_events {schema}.{event_type}[]
) RETURNS SETOF bigint
LANGUAGE plpgsql AS $$
BEGIN
    RETURN QUERY
    WITH new_data AS (
        SELECT * FROM unnest(new_events)
    ),
    inserted AS (
        INSERT INTO {schema}.{events_table} (type, data, tags)
        SELECT type, data, tags
        FROM new_data
        RETURNING id, tags
    ),
    expanded_tags AS (
        SELECT ins.id AS main_id, tag
        FROM inserted ins,
             unnest(ins.tags) AS tag
    ),
    tag_insert AS (
        INSERT INTO {schema}.{tags_table} (tag, main_id)
        SELECT tag, main_id
        FROM expanded_tags
    )
    SELECT MAX(id) FROM inserted;
    NOTIFY {channel};

END
$$;
""")

SQL_CONDITIONAL_APPEND = SQL("""
SELECT * FROM {schema}.{conditional_append}(%(query_items)s, %(after)s, %(events)s)
""")
DB_FUNCTION_NAME_DCB_CONDITIONAL_APPEND_TT = "dcb_conditional_append_tt"
DB_FUNCTION_CONDITIONAL_APPEND = SQL("""
CREATE OR REPLACE FUNCTION {schema}.{conditional_append}(
    query_items {schema}.{query_item_type}[],
    after_id bigint,
    new_events {schema}.{event_type}[]
) RETURNS SETOF bigint
LANGUAGE plpgsql AS $$
DECLARE
    conflict_exists boolean;
BEGIN
    -- Step 0: Lock table in exclusive mode (reads can still read)
    SET LOCAL lock_timeout = '{lock_timeout}s';
    LOCK TABLE {schema}.{events_table} IN EXCLUSIVE MODE;

    -- Step 1: Check for conflicts
    WITH query_items_cte AS (
        SELECT * FROM unnest(query_items) WITH ORDINALITY
    ),
    initial_matches AS (
        SELECT
            t.main_id,
            qi.ordinality,
            t.tag,
            qi.tags AS required_tags,
            qi.types AS allowed_types
        FROM query_items_cte qi
        JOIN {schema}.{tags_table} t
          ON t.tag = ANY(qi.tags)
        WHERE t.main_id > COALESCE(after_id, 0)
    ),
    matched_groups AS (
        SELECT
            main_id,
            ordinality,
            COUNT(DISTINCT tag) AS matched_tag_count,
            array_length(required_tags, 1) AS required_tag_count,
            allowed_types
        FROM initial_matches
        GROUP BY main_id, ordinality, required_tag_count, allowed_types
    ),
    qualified_ids AS (
        SELECT main_id, allowed_types
        FROM matched_groups
        WHERE matched_tag_count = required_tag_count
    ),
    filtered_ids AS (
        SELECT m.id
        FROM {schema}.{events_table} m
        JOIN qualified_ids q ON q.main_id = m.id
        WHERE
            m.id > COALESCE(after_id, 0)
            AND (
                array_length(q.allowed_types, 1) IS NULL
                OR array_length(q.allowed_types, 1) = 0
                OR m.type = ANY(q.allowed_types)
            )
        LIMIT 1
    )
    SELECT EXISTS (SELECT 1 FROM filtered_ids)
    INTO conflict_exists;

    -- Step 2: Insert if no conflicts
    IF NOT conflict_exists THEN
        RETURN QUERY
        WITH new_data AS (
            SELECT * FROM unnest(new_events)
        ),
        inserted AS (
            INSERT INTO {schema}.{events_table} (type, data, tags)
            SELECT type, data, tags
            FROM new_data
            RETURNING id, tags
        ),
        expanded_tags AS (
            SELECT ins.id AS main_id, tag
            FROM inserted ins,
                 unnest(ins.tags) AS tag
        ),
        tag_insert AS (
            INSERT INTO {schema}.{tags_table} (tag, main_id)
            SELECT tag, main_id
            FROM expanded_tags
        )
        SELECT MAX(id) FROM inserted;
        NOTIFY {channel};

    END IF;

    -- If conflict exists, return empty result
    RETURN;
END
$$;
""")


SQL_SET_LOCAL_LOCK_TIMEOUT = SQL("SET LOCAL lock_timeout = '{lock_timeout}s'")
SQL_LOCK_TABLE = SQL("LOCK TABLE {schema}.{events_table} IN EXCLUSIVE MODE")

SQL_EXPLAIN = SQL("EXPLAIN")
SQL_EXPLAIN_ANALYZE = SQL("EXPLAIN (ANALYZE, BUFFERS, VERBOSE)")


class PostgresDCBRecorderTT(DCBRecorder, PostgresRecorder):
    def __init__(
        self,
        datastore: PostgresDatastore,
        *,
        events_table_name: str = "dcb_events",
    ):
        super().__init__(datastore)
        # Define identifiers.
        self.events_table_name = events_table_name + "_tt_main"
        self.channel_name = self.events_table_name.replace(".", "_")
        self.tags_table_name = events_table_name + "_tt_tag"
        self.index_name_id_cover_type = self.events_table_name + "_idx_id_type"
        self.index_name_tag_main_id = self.tags_table_name + "_idx_tag_main_id"

        # Check identifier lengths.
        self.check_identifier_length(self.events_table_name)
        self.check_identifier_length(self.tags_table_name)
        self.check_identifier_length(self.index_name_id_cover_type)
        self.check_identifier_length(self.index_name_tag_main_id)
        self.check_identifier_length(DB_TYPE_NAME_DCB_EVENT_TT)
        self.check_identifier_length(DB_TYPE_NAME_DCB_QUERY_ITEM_TT)

        # Register composite database types.
        self.datastore.db_type_names.add(DB_TYPE_NAME_DCB_EVENT_TT)
        self.datastore.db_type_names.add(DB_TYPE_NAME_DCB_QUERY_ITEM_TT)
        self.datastore.register_type_adapters()

        # Define SQL template keyword arguments.
        self.sql_kwargs = {
            "schema": Identifier(self.datastore.schema),
            "events_table": Identifier(self.events_table_name),
            "channel": Identifier(self.channel_name),
            "tags_table": Identifier(self.tags_table_name),
            "event_type": Identifier(DB_TYPE_NAME_DCB_EVENT_TT),
            "query_item_type": Identifier(DB_TYPE_NAME_DCB_QUERY_ITEM_TT),
            "id_cover_type_index": Identifier(self.index_name_id_cover_type),
            "tag_main_id_index": Identifier(self.index_name_tag_main_id),
            "unconditional_append": Identifier(
                DB_FUNCTION_NAME_DCB_UNCONDITIONAL_APPEND_TT
            ),
            "conditional_append": Identifier(
                DB_FUNCTION_NAME_DCB_CONDITIONAL_APPEND_TT
            ),
            "lock_timeout": self.datastore.lock_timeout,
        }

        # Format and extend SQL create statements.
        self.sql_create_statements.extend(
            [
                self.format(DB_TYPE_DCB_EVENT),
                self.format(DB_TYPE_DCB_QUERY_ITEM),
                self.format(DB_TABLE_DCB_EVENTS),
                self.format(DB_INDEX_UNIQUE_ID_COVER_TYPE),
                self.format(DB_TABLE_DCB_TAGS),
                self.format(DB_INDEX_TAG_MAIN_ID),
                self.format(DB_FUNCTION_UNCONDITIONAL_APPEND),
                self.format(DB_FUNCTION_CONDITIONAL_APPEND),
            ]
        )

        # Format other SQL statements.
        self.sql_select_by_tags = self.format(SQL_SELECT_BY_TAGS)
        self.sql_select_all = self.format(SQL_SELECT_ALL)
        self.sql_select_by_type = self.format(SQL_SELECT_EVENTS_BY_TYPE)
        self.sql_select_max_id = self.format(SQL_SELECT_MAX_ID)
        self.sql_unconditional_append = self.format(SQL_UNCONDITIONAL_APPEND)
        self.sql_conditional_append = self.format(SQL_CONDITIONAL_APPEND)
        self.sql_set_local_lock_timeout = self.format(SQL_SET_LOCAL_LOCK_TIMEOUT)
        self.sql_lock_table = self.format(SQL_LOCK_TABLE)

    def format(self, sql: SQL) -> Composed:
        return sql.format(**self.sql_kwargs)

    def read(
        self,
        query: DCBQuery | None = None,
        *,
        after: int | None = None,
        limit: int | None = None,
    ) -> DCBReadResponse:
        with self.datastore.cursor() as curs:
            events, head = self._read(
                curs=curs,
                query=query,
                after=after,
                limit=limit,
                return_head=True,
            )
            # TODO: Actually return an iterator from _read()!
            return SimpleDCBReadResponse(events=iter(events), head=head)

    def _read(
        self,
        curs: Cursor[DictRow],
        query: DCBQuery | None = None,
        *,
        after: int | None = None,
        limit: int | None = None,
        return_head: bool = True,
    ) -> tuple[Sequence[DCBSequencedEvent], int | None]:
        if return_head and limit is None:
            self.execute(curs, self.sql_select_max_id, explain=False)
            row = curs.fetchone()
            head = None if row is None else row["max"]
        else:
            head = None

        if not query or not query.items:
            # Select all.
            self.execute(
                curs,
                self.sql_select_all,
                {
                    "after": after,
                    "limit": limit,
                },
                explain=False,
            )
            rows = curs.fetchall()

        elif self.all_query_items_have_tags(query):
            # Select with tags.
            psycopg_dcb_query_items = self.construct_psycopg_query_items(query.items)

            self.execute(
                curs,
                self.sql_select_by_tags,
                {
                    "query_items": psycopg_dcb_query_items,
                    "after": after,
                    "limit": limit,
                },
                explain=False,
            )
            rows = curs.fetchall()

        elif self.has_one_query_item_one_type(query):
            # Select for one type.
            self.execute(
                curs,
                self.sql_select_by_type,
                {
                    "event_type": query.items[0].types[0],
                    "after": after,
                    "limit": limit,
                },
                explain=False,
            )
            rows = curs.fetchall()

        else:
            msg = f"Unsupported query: {query}"
            raise ProgrammingError(msg)

        events = [
            DCBSequencedEvent(
                event=DCBEvent(
                    type=row["type"],
                    data=row["data"],
                    tags=row["tags"],
                ),
                position=row["id"],
            )
            for row in rows
        ]

        # Maybe update head.
        if return_head and events:
            head = max(head or 0, *[e.position for e in events])

        return events, head

    def subscribe(
        self,
        query: DCBQuery | None = None,
        *,
        after: int | None = None,
    ) -> PostgresDCBSubscription:
        return PostgresDCBSubscription(
            recorder=self,
            query=query,
            after=after,
        )

    def append(
        self, events: Sequence[DCBEvent], condition: DCBAppendCondition | None = None
    ) -> int:
        assert len(events) > 0
        psycopg_dcb_events = self.construct_psycopg_dcb_events(events)

        # Do single-statement "unconditional append".
        if condition is None:
            with self.datastore.cursor() as curs:
                return self._unconditional_append(curs, psycopg_dcb_events)

        if self.all_query_items_have_tags(condition.fail_if_events_match):
            # Do single-statement "conditional append".
            psycopg_dcb_query_items = self.construct_psycopg_query_items(
                condition.fail_if_events_match.items
            )
            with self.datastore.cursor() as curs:
                self.execute(
                    curs,
                    self.sql_conditional_append,
                    {
                        "query_items": psycopg_dcb_query_items,
                        "after": condition.after,
                        "events": psycopg_dcb_events,
                    },
                    explain=False,
                )
                row = curs.fetchone()
                if row is None:
                    raise IntegrityError

                return row[DB_FUNCTION_NAME_DCB_CONDITIONAL_APPEND_TT]

        # Do separate "read" and "append" operations in a transaction.
        with self.datastore.transaction(commit=True) as curs:

            # Lock the table in exclusive mode (readers can still read) to ensure
            # nothing else will execute an append condition statement until after
            # we have finished inserting new events, whilst expecting that others
            # are playing by the same game. By the way, this is how optimistic
            # locking works.
            if self.datastore.lock_timeout:
                curs.execute(self.sql_set_local_lock_timeout)
            curs.execute(self.sql_lock_table)

            # Check the append condition.
            failed, _ = self._read(
                curs=curs,
                query=condition.fail_if_events_match,
                after=condition.after,
                limit=1,
                return_head=False,
            )
            if failed:
                raise IntegrityError(failed)

            # If okay, then do an "unconditional append".
            return self._unconditional_append(curs, psycopg_dcb_events)

    def _unconditional_append(
        self, curs: Cursor[DictRow], psycopg_dcb_events: list[PsycopgDCBEvent]
    ) -> int:
        self.execute(
            curs,
            self.sql_unconditional_append,
            {
                "events": psycopg_dcb_events,
            },
            explain=False,
        )
        row = curs.fetchone()
        if row is None:  # pragma: no cover
            msg = "Shouldn't get here"
            raise InternalError(msg)

        return row[DB_FUNCTION_NAME_DCB_UNCONDITIONAL_APPEND_TT]

    def construct_psycopg_dcb_events(
        self, dcb_events: Sequence[DCBEvent]
    ) -> list[PsycopgDCBEvent]:
        return [
            self.datastore.psycopg_python_types[DB_TYPE_NAME_DCB_EVENT_TT](
                type=e.type,
                data=e.data,
                tags=e.tags,
            )
            for e in dcb_events
        ]

    def construct_psycopg_query_items(
        self, query_items: Sequence[DCBQueryItem]
    ) -> list[PsycopgDCBQueryItem]:
        return [
            self.datastore.psycopg_python_types[DB_TYPE_NAME_DCB_QUERY_ITEM_TT](
                types=q.types,
                tags=q.tags,
            )
            for q in query_items
        ]

    def has_one_query_item_one_type(self, query: DCBQuery) -> bool:
        return (
            len(query.items) == 1
            and len(query.items[0].types) == 1
            and len(query.items[0].tags) == 0
        )

    def all_query_items_have_tags(self, query: DCBQuery) -> bool:
        return all(len(q.tags) > 0 for q in query.items) and len(query.items) > 0

    def execute(
        self,
        cursor: Cursor[DictRow],
        statement: Composed,
        params: Params | None = None,
        *,
        explain: bool = False,
        prepare: bool = True,
    ) -> None:
        if explain:  # pragma: no cover
            print()  # noqa: T201
            print("Statement:", statement.as_string().strip())  # noqa: T201
            print("Params:", params)  # noqa: T201
            print()  # noqa: T201
            # with self.datastore.transaction(commit=False) as explain_cursor:
            #     explain_cursor.execute(SQL_EXPLAIN + statement, params)
            #     rows = explain_cursor.fetchall()
            #     print("\n".join([r["QUERY PLAN"] for r in rows]))  # no qa: T201
            #     print()  # no qa: T201
            # with self.datastore.transaction(commit=False) as explain_cursor:
            cursor.execute(SQL_EXPLAIN + statement, params)
            rows = cursor.fetchall()
            print("\n".join([r["QUERY PLAN"] for r in rows]))  # noqa: T201
            print()  # noqa: T201
        cursor.execute(statement, params, prepare=prepare)


class PsycopgDCBEvent(NamedTuple):
    type: str
    data: bytes
    tags: list[str]


class PsycopgDCBQueryItem(NamedTuple):
    types: list[str]
    tags: list[str]


class PostgresDCBSubscription(DCBListenNotifySubscription[PostgresDCBRecorderTT]):
    def __init__(
        self,
        recorder: PostgresDCBRecorderTT,
        query: DCBQuery | None = None,
        after: int | None = None,
    ) -> None:
        super().__init__(recorder=recorder, query=query, after=after)
        self._has_listen_connection = threading.Event()
        self._listen_connection: Connection[dict[str, Any]] | None = None
        self._listen_thread = threading.Thread(target=self._listen, daemon=True)
        self._listen_thread.start()

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        super().__exit__(*args, **kwargs)
        self._listen_thread.join()

    def _listen(self) -> None:
        recorder = self._recorder
        assert isinstance(recorder, PostgresDCBRecorderTT)
        try:
            with recorder.datastore.get_connection() as conn:
                self._listen_connection = conn
                self._has_listen_connection.set()
                conn.execute(
                    SQL("LISTEN {0}").format(Identifier(recorder.channel_name))
                )
                while not self._has_been_stopped and not self._thread_error:
                    # This block simplifies psycopg's conn.notifies(), because
                    # we aren't interested in the actual notify messages, and
                    # also we want to stop consuming notify messages when the
                    # subscription has an error or is otherwise stopped.
                    with conn.lock:
                        try:
                            if conn.wait(notifies(conn.pgconn), interval=1):
                                self._has_been_notified.set()
                        except NO_TRACEBACK as ex:  # pragma: no cover
                            raise ex.with_traceback(None) from None

        except BaseException as e:  # pragma: no cover
            if self._thread_error is None:
                self._thread_error = e
            self.stop()


class PostgresTTDCBFactory(
    BasePostgresFactory[PostgresTrackingRecorder],
    DCBInfrastructureFactory[PostgresTrackingRecorder],
):
    def dcb_recorder(self) -> DCBRecorder:
        prefix = self.env.name.lower() or "dcb"

        dcb_table_name = prefix + "_events"
        recorder = PostgresDCBRecorderTT(
            datastore=self.datastore,
            events_table_name=dcb_table_name,
        )
        if self.env_create_table():
            recorder.create_table()
        return recorder



from __future__ import annotations

import threading
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Generic

from eventsourcing.dcb.api import (
    DCBAppendCondition,
    DCBEvent,
    DCBQuery,
    DCBQueryItem,
    DCBReadResponse,
    DCBRecorder,
    DCBSequencedEvent,
    DCBSubscription,
    TDCBRecorder_co,
)
from eventsourcing.dcb.domain import (
    Selector,
    Tagged,
    TDecision,
)
from eventsourcing.persistence import (
    BaseInfrastructureFactory,
    Queue,
    ShutDown,
    TTrackingRecorder,
)
from eventsourcing.utils import get_topic

if TYPE_CHECKING:
    from collections.abc import Sequence


class DCBMapper(ABC, Generic[TDecision]):
    @abstractmethod
    def to_dcb_event(self, event: Tagged[TDecision]) -> DCBEvent:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def to_domain_event(self, event: DCBEvent) -> Tagged[TDecision]:
        raise NotImplementedError  # pragma: no cover


class DCBEventStore(Generic[TDecision]):
    def __init__(self, mapper: DCBMapper[TDecision], recorder: DCBRecorder):
        self.mapper = mapper
        self.recorder = recorder

    def append(
        self,
        events: Sequence[Tagged[TDecision]],
        cb: Selector | Sequence[Selector] | None = None,
        after: int | None = None,
    ) -> int:
        if len(events) == 0:
            return 0
        condition = (
            None
            if cb is None and after is None
            else DCBAppendCondition(
                fail_if_events_match=self._cb_to_dcb_query(cb),
                after=after,
            )
        )
        return self.recorder.append(
            events=[self.mapper.to_dcb_event(e) for e in events],
            condition=condition,
        )

    def read(
        self,
        cb: Selector | Sequence[Selector] | None = None,
        *,
        after: int | None = None,
    ) -> DCBEventStoreReadResponse[TDecision]:
        query = self._cb_to_dcb_query(cb)
        read_response = self.recorder.read(
            query=query,
            after=after,
        )
        return DCBEventStoreReadResponse(read_response, self.mapper)

    @staticmethod
    def _cb_to_dcb_query(
        cb: Selector | Sequence[Selector] | None = None,
    ) -> DCBQuery:
        cb = [cb] if isinstance(cb, Selector) else cb or []
        return DCBQuery(
            items=[
                DCBQueryItem(
                    types=[get_topic(t) for t in s.types],
                    tags=list(s.tags),
                )
                for s in cb
            ]
        )


class DCBEventStoreReadResponse(Iterator[Tagged[TDecision]]):
    def __init__(
        self, dcb_read_response: DCBReadResponse, mapper: DCBMapper[TDecision]
    ):
        self._dcb_read_response = dcb_read_response
        self._mapper = mapper

    @property
    def head(self) -> int | None:
        return self._dcb_read_response.head

    def __next__(self) -> Tagged[TDecision]:
        dcb_sequenced_event = self._dcb_read_response.__next__()
        return self._mapper.to_domain_event(dcb_sequenced_event.event)


class NotFoundError(Exception):
    pass


class DCBInfrastructureFactory(BaseInfrastructureFactory[TTrackingRecorder], ABC):
    @abstractmethod
    def dcb_recorder(self) -> DCBRecorder:
        pass  # pragma: no cover


class DCBListenNotifySubscription(DCBSubscription[TDCBRecorder_co]):
    def __init__(
        self,
        recorder: TDCBRecorder_co,
        query: DCBQuery | None = None,
        after: int | None = None,
    ) -> None:
        super().__init__(recorder=recorder, query=query, after=after)
        self.select_limit = 500
        self._events: Sequence[DCBSequencedEvent] = []
        self._events_index: int = 0
        self._events_queue: Queue[Sequence[DCBSequencedEvent]] = Queue(maxsize=10)
        self._has_been_notified = threading.Event()
        self._thread_error: BaseException | None = None
        self._pull_thread = threading.Thread(target=self._loop_on_pull)
        self._pull_thread.start()

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        super().__exit__(*args, **kwargs)
        self._pull_thread.join()

    def stop(self) -> None:
        """Stops the subscription."""
        super().stop()
        self._events_queue.shutdown(  # pyright: ignore[reportAttributeAccessIssue]
            immediate=True
        )
        self._has_been_notified.set()

    def __next__(self) -> DCBSequencedEvent:
        # If necessary, get a new list of events from the recorder.
        if self._events_index == len(self._events) and not self._has_been_stopped:
            try:
                self._events = self._events_queue.get()
            except ShutDown:
                pass
            else:
                self._events_queue.task_done()
                self._events_index = 0

        # Stop the iteration if subscription has been stopped.
        if self._has_been_stopped:
            # Maybe raise thread error.
            if self._thread_error is not None:
                raise self._thread_error
            raise StopIteration

        # Return an event from previously obtained list.
        event = self._events[self._events_index]
        self._events_index += 1
        return event

    def _loop_on_pull(self) -> None:
        try:
            self._pull()  # Already recorded events.
            while not self._has_been_stopped:
                self._has_been_notified.wait()
                self._pull()  # Newly recorded events.
        except BaseException as e:  # pragma: no cover
            if self._thread_error is None:
                self._thread_error = e
            self.stop()

    def _pull(self) -> None:
        while not self._has_been_stopped:
            self._has_been_notified.clear()
            events = list(
                self._recorder.read(
                    query=self._query,
                    after=self._last_position,
                    limit=self.select_limit,
                )
            )
            if len(events) > 0:
                try:
                    self._events_queue.put(events)
                except ShutDown:  # pragma: no cover
                    # TODO: Cover this with a test...
                    break
                self._last_position = events[-1].position
            if len(events) < self.select_limit:
                break



from __future__ import annotations

import traceback
import zlib
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from tempfile import NamedTemporaryFile
from threading import Event, Thread, get_ident
from time import sleep
from timeit import timeit
from typing import TYPE_CHECKING, Any, Generic, cast
from unittest import TestCase
from uuid import UUID, uuid4

from typing_extensions import TypeVar

from eventsourcing.cipher import AESCipher
from eventsourcing.compressor import ZlibCompressor
from eventsourcing.domain import DomainEvent, datetime_now_with_tzinfo
from eventsourcing.persistence import (
    AggregateRecorder,
    ApplicationRecorder,
    DatetimeAsISO,
    DecimalAsStr,
    InfrastructureFactory,
    IntegrityError,
    JSONTranscoder,
    Mapper,
    Notification,
    ProcessRecorder,
    StoredEvent,
    Tracking,
    TrackingRecorder,
    Transcoder,
    Transcoding,
    UUIDAsHex,
    WaitInterruptedError,
)
from eventsourcing.utils import Environment, get_topic

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from typing_extensions import Never


class RecorderTestCase(TestCase, ABC):
    INITIAL_VERSION = 1

    def new_originator_id(self) -> UUID | str:
        return uuid4()


class AggregateRecorderTestCase(RecorderTestCase, ABC):
    @abstractmethod
    def create_recorder(self) -> AggregateRecorder:
        """"""

    def test_insert_and_select(self) -> None:
        # Construct the recorder.
        recorder = self.create_recorder()

        # Check we can call insert_events() with an empty list.
        notification_ids = recorder.insert_events([])
        self.assertEqual(notification_ids, None)

        # Select stored events, expect empty list.
        originator_id1 = self.new_originator_id()
        self.assertEqual(
            recorder.select_events(originator_id1, desc=True, limit=1),
            [],
        )

        # Write a stored event.
        stored_event1 = StoredEvent(
            originator_id=originator_id1,
            originator_version=self.INITIAL_VERSION,
            topic="topic1",
            state=b"state1",
        )
        notification_ids = recorder.insert_events([stored_event1])
        self.assertEqual(notification_ids, None)

        # Select stored events, expect list of one.
        stored_events = recorder.select_events(originator_id1)
        self.assertEqual(len(stored_events), 1)
        self.assertEqual(stored_events[0].originator_id, originator_id1)
        self.assertEqual(stored_events[0].originator_version, self.INITIAL_VERSION)
        self.assertEqual(stored_events[0].topic, "topic1")
        self.assertEqual(stored_events[0].state, b"state1")
        self.assertIsInstance(stored_events[0].state, bytes)

        # Check get record conflict error if attempt to store it again.
        with self.assertRaises(IntegrityError):
            recorder.insert_events([stored_event1])

        # Check writing of events is atomic.
        stored_event2 = StoredEvent(
            originator_id=originator_id1,
            originator_version=self.INITIAL_VERSION + 1,
            topic="topic2",
            state=b"state2",
        )
        with self.assertRaises(IntegrityError):
            recorder.insert_events([stored_event1, stored_event2])

        with self.assertRaises(IntegrityError):
            recorder.insert_events([stored_event2, stored_event2])

        # Check still only have one record.
        stored_events = recorder.select_events(originator_id1)
        self.assertEqual(len(stored_events), 1)
        self.assertEqual(stored_events[0].originator_id, stored_event1.originator_id)
        self.assertEqual(
            stored_events[0].originator_version, stored_event1.originator_version
        )
        self.assertEqual(stored_events[0].topic, stored_event1.topic)

        # Check can write two events together.
        stored_event3 = StoredEvent(
            originator_id=originator_id1,
            originator_version=self.INITIAL_VERSION + 2,
            topic="topic3",
            state=b"state3",
        )
        notification_ids = recorder.insert_events([stored_event2, stored_event3])
        self.assertEqual(notification_ids, None)

        # Check we got what was written.
        stored_events = recorder.select_events(originator_id1)
        self.assertEqual(len(stored_events), 3)
        self.assertEqual(stored_events[0].originator_id, originator_id1)
        self.assertEqual(stored_events[0].originator_version, self.INITIAL_VERSION)
        self.assertEqual(stored_events[0].topic, "topic1")
        self.assertEqual(stored_events[0].state, b"state1")
        self.assertEqual(stored_events[1].originator_id, originator_id1)
        self.assertEqual(stored_events[1].originator_version, self.INITIAL_VERSION + 1)
        self.assertEqual(stored_events[1].topic, "topic2")
        self.assertEqual(stored_events[1].state, b"state2")
        self.assertEqual(stored_events[2].originator_id, originator_id1)
        self.assertEqual(stored_events[2].originator_version, self.INITIAL_VERSION + 2)
        self.assertEqual(stored_events[2].topic, "topic3")
        self.assertEqual(stored_events[2].state, b"state3")

        # Check we can get the last one recorded (used to get last snapshot).
        stored_events = recorder.select_events(originator_id1, desc=True, limit=1)
        self.assertEqual(len(stored_events), 1)
        self.assertEqual(
            stored_events[0],
            stored_event3,
        )

        # Check we can get the last one before a particular version.
        stored_events = recorder.select_events(
            originator_id1, lte=self.INITIAL_VERSION + 1, desc=True, limit=1
        )
        self.assertEqual(len(stored_events), 1)
        self.assertEqual(
            stored_events[0],
            stored_event2,
        )

        # Check we can get events between versions (historical state with snapshot).
        stored_events = recorder.select_events(
            originator_id1, gt=self.INITIAL_VERSION, lte=self.INITIAL_VERSION + 1
        )
        self.assertEqual(len(stored_events), 1)
        self.assertEqual(
            stored_events[0],
            stored_event2,
        )

        # Check aggregate sequences are distinguished.
        originator_id2 = self.new_originator_id()
        self.assertEqual(
            recorder.select_events(originator_id2),
            [],
        )

        # Write a stored event in a different sequence.
        stored_event4 = StoredEvent(
            originator_id=originator_id2,
            originator_version=0,
            topic="topic4",
            state=b"state4",
        )
        recorder.insert_events([stored_event4])
        stored_events = recorder.select_events(originator_id2)
        self.assertEqual(
            stored_events,
            [stored_event4],
        )

    def test_performance(self) -> None:
        # Construct the recorder.
        recorder = self.create_recorder()

        def insert() -> None:
            originator_id = self.new_originator_id()

            stored_event = StoredEvent(
                originator_id=originator_id,
                originator_version=self.INITIAL_VERSION,
                topic="topic1",
                state=b"state1",
            )
            recorder.insert_events([stored_event])

        # Warm up.
        number = 10
        timeit(insert, number=number)

        number = 100
        duration = timeit(insert, number=number)
        print(
            self,
            f"\n{1000000 * duration / number:.1f} μs per insert, "
            f"{number / duration:.0f} inserts per second",
        )


_TApplicationRecorder = TypeVar(
    "_TApplicationRecorder", bound=ApplicationRecorder, default=ApplicationRecorder
)


class ApplicationRecorderTestCase(
    RecorderTestCase, ABC, Generic[_TApplicationRecorder]
):
    EXPECT_CONTIGUOUS_NOTIFICATION_IDS = True

    @abstractmethod
    def create_recorder(self) -> _TApplicationRecorder:
        """"""

    def test_insert_select(self) -> None:
        # Construct the recorder.
        recorder = self.create_recorder()

        # Check notifications methods work when there aren't any.
        self.assertEqual(len(recorder.select_notifications(start=None, limit=3)), 0)
        self.assertEqual(
            len(recorder.select_notifications(start=None, limit=3, topics=["topic1"])),
            0,
        )

        self.assertIsNone(recorder.max_notification_id())

        # Write two stored events.
        originator_id1 = self.new_originator_id()
        originator_id2 = self.new_originator_id()

        stored_event1 = StoredEvent(
            originator_id=originator_id1,
            originator_version=self.INITIAL_VERSION,
            topic="topic1",
            state=b"state1",
        )
        stored_event2 = StoredEvent(
            originator_id=originator_id1,
            originator_version=self.INITIAL_VERSION + 1,
            topic="topic2",
            state=b"state2",
        )

        notification_ids = recorder.insert_events([stored_event1, stored_event2])
        self.assertEqual(notification_ids, [1, 2])

        # Store a third event.
        stored_event3 = StoredEvent(
            originator_id=originator_id2,
            originator_version=self.INITIAL_VERSION,
            topic="topic3",
            state=b"state3",
        )
        notification_ids = recorder.insert_events([stored_event3])
        self.assertEqual(notification_ids, [3])

        stored_events1 = recorder.select_events(originator_id1)
        stored_events2 = recorder.select_events(originator_id2)

        # Check we got what was written.
        self.assertEqual(len(stored_events1), 2)
        self.assertEqual(len(stored_events2), 1)

        # Check get record conflict error if attempt to store it again.
        with self.assertRaises(IntegrityError):
            recorder.insert_events([stored_event3])

        # sleep(1)  # Added to make eventsourcing-axon tests work.
        notifications = recorder.select_notifications(start=None, limit=10)
        self.assertEqual(len(notifications), 3)
        self.assertEqual(notifications[0].id, 1)
        self.assertEqual(notifications[0].originator_id, originator_id1)
        self.assertEqual(notifications[0].topic, "topic1")
        self.assertEqual(notifications[0].state, b"state1")
        self.assertEqual(notifications[1].id, 2)
        self.assertEqual(notifications[1].originator_id, originator_id1)
        self.assertEqual(notifications[1].topic, "topic2")
        self.assertEqual(notifications[1].state, b"state2")
        self.assertEqual(notifications[2].id, 3)
        self.assertEqual(notifications[2].originator_id, originator_id2)
        self.assertEqual(notifications[2].topic, "topic3")
        self.assertEqual(notifications[2].state, b"state3")

        notifications = recorder.select_notifications(start=1, limit=10)
        self.assertEqual(len(notifications), 3)
        self.assertEqual(notifications[0].id, 1)
        self.assertEqual(notifications[0].originator_id, originator_id1)
        self.assertEqual(notifications[0].topic, "topic1")
        self.assertEqual(notifications[0].state, b"state1")
        self.assertEqual(notifications[1].id, 2)
        self.assertEqual(notifications[1].originator_id, originator_id1)
        self.assertEqual(notifications[1].topic, "topic2")
        self.assertEqual(notifications[1].state, b"state2")
        self.assertEqual(notifications[2].id, 3)
        self.assertEqual(notifications[2].originator_id, originator_id2)
        self.assertEqual(notifications[2].topic, "topic3")
        self.assertEqual(notifications[2].state, b"state3")

        notifications = recorder.select_notifications(start=None, stop=2, limit=10)
        self.assertEqual(len(notifications), 2)
        self.assertEqual(notifications[0].id, 1)
        self.assertEqual(notifications[0].originator_id, originator_id1)
        self.assertEqual(notifications[0].topic, "topic1")
        self.assertEqual(notifications[0].state, b"state1")
        self.assertEqual(notifications[1].id, 2)
        self.assertEqual(notifications[1].originator_id, originator_id1)
        self.assertEqual(notifications[1].topic, "topic2")
        self.assertEqual(notifications[1].state, b"state2")

        notifications = recorder.select_notifications(
            start=1, limit=10, inclusive_of_start=False
        )
        self.assertEqual(len(notifications), 2)
        self.assertEqual(notifications[0].id, 2)
        self.assertEqual(notifications[0].originator_id, originator_id1)
        self.assertEqual(notifications[0].topic, "topic2")
        self.assertEqual(notifications[0].state, b"state2")
        self.assertEqual(notifications[1].id, 3)
        self.assertEqual(notifications[1].originator_id, originator_id2)
        self.assertEqual(notifications[1].topic, "topic3")
        self.assertEqual(notifications[1].state, b"state3")

        notifications = recorder.select_notifications(
            start=2, limit=10, inclusive_of_start=False
        )
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].id, 3)
        self.assertEqual(notifications[0].originator_id, originator_id2)
        self.assertEqual(notifications[0].topic, "topic3")
        self.assertEqual(notifications[0].state, b"state3")

        notifications = recorder.select_notifications(
            start=None, limit=10, topics=["topic1", "topic2", "topic3"]
        )
        self.assertEqual(len(notifications), 3)
        self.assertEqual(notifications[0].id, 1)
        self.assertEqual(notifications[0].originator_id, originator_id1)
        self.assertEqual(notifications[0].topic, "topic1")
        self.assertEqual(notifications[0].state, b"state1")
        self.assertEqual(notifications[1].id, 2)
        self.assertEqual(notifications[1].originator_id, originator_id1)
        self.assertEqual(notifications[1].topic, "topic2")
        self.assertEqual(notifications[1].state, b"state2")
        self.assertEqual(notifications[2].id, 3)
        self.assertEqual(notifications[2].originator_id, originator_id2)
        self.assertEqual(notifications[2].topic, "topic3")
        self.assertEqual(notifications[2].state, b"state3")

        notifications = recorder.select_notifications(1, 10, topics=["topic1"])
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].id, 1)
        self.assertEqual(notifications[0].originator_id, originator_id1)
        self.assertEqual(notifications[0].topic, "topic1")
        self.assertEqual(notifications[0].state, b"state1")

        notifications = recorder.select_notifications(1, 3, topics=["topic2"])
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].id, 2)
        self.assertEqual(notifications[0].originator_id, originator_id1)
        self.assertEqual(notifications[0].topic, "topic2")
        self.assertEqual(notifications[0].state, b"state2")

        notifications = recorder.select_notifications(1, 3, topics=["topic3"])
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].id, 3)
        self.assertEqual(notifications[0].originator_id, originator_id2)
        self.assertEqual(notifications[0].topic, "topic3")
        self.assertEqual(notifications[0].state, b"state3")

        notifications = recorder.select_notifications(1, 3, topics=["topic1", "topic3"])
        self.assertEqual(len(notifications), 2)
        self.assertEqual(notifications[0].id, 1)
        self.assertEqual(notifications[0].originator_id, originator_id1)
        self.assertEqual(notifications[0].topic, "topic1")
        self.assertEqual(notifications[0].state, b"state1")
        self.assertEqual(notifications[1].id, 3)
        self.assertEqual(notifications[1].topic, "topic3")
        self.assertEqual(notifications[1].state, b"state3")

        self.assertEqual(recorder.max_notification_id(), 3)

        # Check limit is working
        notifications = recorder.select_notifications(None, 1)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].id, 1)

        notifications = recorder.select_notifications(2, 1)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].id, 2)

        notifications = recorder.select_notifications(1, 1, inclusive_of_start=False)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].id, 2)

        notifications = recorder.select_notifications(2, 2)
        self.assertEqual(len(notifications), 2)
        self.assertEqual(notifications[0].id, 2)
        self.assertEqual(notifications[1].id, 3)

        notifications = recorder.select_notifications(3, 1)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].id, 3)

        notifications = recorder.select_notifications(3, 1, inclusive_of_start=False)
        self.assertEqual(len(notifications), 0)

        notifications = recorder.select_notifications(start=2, limit=10, stop=2)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0].id, 2)

        notifications = recorder.select_notifications(start=1, limit=10, stop=2)
        self.assertEqual(len(notifications), 2, len(notifications))
        self.assertEqual(notifications[0].id, 1)
        self.assertEqual(notifications[1].id, 2)

        notifications = recorder.select_notifications(
            start=1, limit=10, stop=2, inclusive_of_start=False
        )
        self.assertEqual(len(notifications), 1, len(notifications))
        self.assertEqual(notifications[0].id, 2)

    def test_performance(self) -> None:
        # Construct the recorder.
        recorder = self.create_recorder()

        def insert() -> None:
            originator_id = self.new_originator_id()

            stored_event = StoredEvent(
                originator_id=originator_id,
                originator_version=self.INITIAL_VERSION,
                topic="topic1",
                state=b"state1",
            )
            recorder.insert_events([stored_event])

        # Warm up.
        number = 10
        timeit(insert, number=number)

        number = 100
        duration = timeit(insert, number=number)
        print(
            self,
            f"\n{1000000 * duration / number:.1f} μs per insert, "
            f"{number / duration:.0f} inserts per second",
        )

    def test_concurrent_no_conflicts(self, initial_position: int = 0) -> None:
        print(self)

        recorder = self.create_recorder()

        errors_happened = Event()
        errors: list[Exception] = []

        counts = {}
        threads: dict[int, int] = {}
        durations: dict[int, float] = {}

        num_writers = 10
        num_writes_per_writer = 100
        num_events_per_write = 100
        reader_sleep = 0.0001
        writer_sleep = 0.0001

        def insert_events() -> None:
            thread_id = get_ident()
            if thread_id not in threads:
                threads[thread_id] = len(threads)
            if thread_id not in counts:
                counts[thread_id] = 0
            if thread_id not in durations:
                durations[thread_id] = 0

            # thread_num = threads[thread_id]
            # count = counts[thread_id]

            originator_id = self.new_originator_id()
            stored_events = [
                StoredEvent(
                    originator_id=originator_id,
                    originator_version=i,
                    topic="topic",
                    state=b"state",
                )
                for i in range(num_events_per_write)
            ]
            started = datetime_now_with_tzinfo()
            # print(f"Thread {thread_num} write beginning #{count + 1}")
            try:
                recorder.insert_events(stored_events)

            except Exception as e:  # pragma: no cover
                if errors:
                    return
                ended = datetime_now_with_tzinfo()
                duration = (ended - started).total_seconds()
                print(f"Error after starting {duration}", e)
                errors.append(e)
            else:
                ended = datetime_now_with_tzinfo()
                duration = (ended - started).total_seconds()
                counts[thread_id] += 1
                durations[thread_id] = max(durations[thread_id], duration)
                sleep(writer_sleep)

        stop_reading = Event()

        def read_continuously() -> None:
            while not stop_reading.is_set():
                try:
                    recorder.select_notifications(
                        start=initial_position, limit=10, inclusive_of_start=False
                    )
                except Exception as e:  # pragma: no cover
                    errors.append(e)
                    return
                # else:
                sleep(reader_sleep)

        reader_thread1 = Thread(target=read_continuously)
        reader_thread1.start()

        reader_thread2 = Thread(target=read_continuously)
        reader_thread2.start()

        with ThreadPoolExecutor(max_workers=num_writers) as executor:
            futures = []
            for _ in range(num_writes_per_writer):
                if errors:  # pragma: no cover
                    break
                future = executor.submit(insert_events)
                futures.append(future)
            for future in futures:
                if errors:  # pragma: no cover
                    break
                try:
                    future.result()
                except Exception as e:  # pragma: no cover
                    errors.append(e)
                    break

        stop_reading.set()

        if errors:  # pragma: no cover
            raise errors[0]

        for thread_id, thread_num in threads.items():
            count = counts[thread_id]
            duration = durations[thread_id]
            print(f"Thread {thread_num} wrote {count} times (max dur {duration})")
        self.assertFalse(errors_happened.is_set())

    def test_concurrent_throughput(self) -> None:
        print(self)

        recorder = self.create_recorder()

        errors_happened = Event()

        # Match this to the batch page size in postgres insert for max throughput.
        num_events_per_job = 50
        num_jobs = 10
        num_workers = 4

        def insert_events() -> None:
            originator_id = self.new_originator_id()
            stored_events = [
                StoredEvent(
                    originator_id=originator_id,
                    originator_version=i,
                    topic="topic",
                    state=b"state",
                )
                for i in range(num_events_per_job)
            ]

            try:
                recorder.insert_events(stored_events)

            except Exception:  # pragma: no cover
                errors_happened.set()
                tb = traceback.format_exc()
                print(tb)

        # Warm up.
        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            futures = []
            for _ in range(num_workers):
                future = executor.submit(insert_events)
                futures.append(future)
            for future in futures:
                future.result()

        # Run.
        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            started = datetime_now_with_tzinfo()
            futures = []
            for _ in range(num_jobs):
                future = executor.submit(insert_events)
                futures.append(future)
            for future in futures:
                future.result()

        self.assertFalse(errors_happened.is_set(), "There were errors (see above)")
        ended = datetime_now_with_tzinfo()
        rate = num_jobs * num_events_per_job / (ended - started).total_seconds()
        print(f"Rate: {rate:.0f} inserts per second")

    def optional_test_insert_subscribe(self, initial_position: int = 0) -> None:

        recorder = self.create_recorder()

        # Get the max notification ID.
        max_notification_id1 = recorder.max_notification_id()

        # Write two stored events.
        originator_id1 = self.new_originator_id()
        originator_id2 = self.new_originator_id()

        stored_event1 = StoredEvent(
            originator_id=originator_id1,
            originator_version=self.INITIAL_VERSION,
            topic="topic1",
            state=b"state1",
        )
        stored_event2 = StoredEvent(
            originator_id=originator_id1,
            originator_version=self.INITIAL_VERSION + 1,
            topic="topic2",
            state=b"state2",
        )

        notification_ids = recorder.insert_events([stored_event1, stored_event2])
        if self.EXPECT_CONTIGUOUS_NOTIFICATION_IDS:
            self.assertEqual(
                notification_ids, [1 + initial_position, 2 + initial_position]
            )

        # Get the max notification ID.
        max_notification_id2 = recorder.max_notification_id()

        # Start a subscription with default value for 'start'.
        with recorder.subscribe(gt=initial_position) as subscription:

            # Receive events from the subscription.
            for _ in subscription:
                break

        # Start a subscription with None value for 'start'.
        with recorder.subscribe(gt=initial_position) as subscription:

            # Receive events from the subscription.
            for _ in subscription:
                break

        # Start a subscription with int value for 'start'.
        with recorder.subscribe(gt=max_notification_id1) as subscription:

            # Receive events from the subscription.
            notifications: list[Notification] = []
            for notification in subscription:
                notifications.append(notification)
                if len(notifications) == 2:
                    break

            # Check the events we received are the ones that were written.
            self.assertEqual(
                stored_event1.originator_id, notifications[0].originator_id
            )
            self.assertEqual(
                stored_event1.originator_version, notifications[0].originator_version
            )
            self.assertEqual(
                stored_event2.originator_id, notifications[1].originator_id
            )
            self.assertEqual(
                stored_event2.originator_version, notifications[1].originator_version
            )
            if self.EXPECT_CONTIGUOUS_NOTIFICATION_IDS:
                self.assertEqual(1 + initial_position, notifications[0].id)
                self.assertEqual(2 + initial_position, notifications[1].id)

            # Store a third event.
            stored_event3 = StoredEvent(
                originator_id=originator_id2,
                originator_version=self.INITIAL_VERSION,
                topic="topic3",
                state=b"state3",
            )
            notification_ids = recorder.insert_events([stored_event3])
            if self.EXPECT_CONTIGUOUS_NOTIFICATION_IDS:
                self.assertEqual(notification_ids, [3 + initial_position])

            # Receive events from the subscription.
            for notification in subscription:
                notifications.append(notification)
                if len(notifications) == 3:
                    break

            # Check the events we received are the ones that were written.
            self.assertEqual(
                stored_event3.originator_id, notifications[2].originator_id
            )
            self.assertEqual(
                stored_event3.originator_version, notifications[2].originator_version
            )
            if self.EXPECT_CONTIGUOUS_NOTIFICATION_IDS:
                self.assertEqual(3 + initial_position, notifications[2].id)

        # Start a subscription with int value for 'start'.
        with recorder.subscribe(gt=max_notification_id2) as subscription:

            # Receive events from the subscription.
            notifications = []
            for notification in subscription:
                notifications.append(notification)
                if len(notifications) == 1:
                    break

            # Check the events we received are the ones that were written.
            self.assertEqual(
                stored_event3.originator_id, notifications[0].originator_id
            )

        # Start a subscription, call stop() during iteration.
        with recorder.subscribe(gt=initial_position) as subscription:

            # Receive events from the subscription.
            for i, _ in enumerate(subscription):
                subscription.stop()
                # Shouldn't get here twice...
                self.assertLess(i, 1, "Got here twice")

        # Start a subscription, call stop() before iteration.
        subscription = recorder.subscribe(gt=None)
        with subscription:
            subscription.stop()
            # Receive events from the subscription.
            for _ in subscription:
                # Shouldn't get here...
                self.fail("Got here")

        # Start a subscription, call stop() before entering context manager.
        subscription = recorder.subscribe(gt=None)
        subscription.stop()
        with subscription:
            # Receive events from the subscription.
            for _ in subscription:
                # Shouldn't get here...
                self.fail("Got here")

        # Start a subscription with topics.
        subscription = recorder.subscribe(gt=None, topics=["topic3"])
        with subscription:
            for notification in subscription:
                self.assertEqual(notification.topic, "topic3")
                if (
                    notification.originator_id == stored_event3.originator_id
                    and notification.originator_version
                    == stored_event3.originator_version
                ):
                    break

    def close_db_connection(self, *args: Any) -> None:
        """"""


class TrackingRecorderTestCase(TestCase, ABC):
    @abstractmethod
    def create_recorder(self) -> TrackingRecorder:
        """"""

    def test_insert_tracking(self) -> None:
        tracking_recorder = self.create_recorder()

        # Construct tracking objects.
        tracking1 = Tracking("upstream1", 21)
        tracking2 = Tracking("upstream1", 22)
        tracking3 = Tracking("upstream2", 21)

        # Insert tracking objects.
        tracking_recorder.insert_tracking(tracking=tracking1)
        tracking_recorder.insert_tracking(tracking=tracking2)
        tracking_recorder.insert_tracking(tracking=tracking3)

        # raise Exception(tracking_recorder.max_tracking_id(tracking1.application_name))

        # Fail to insert same tracking object twice.
        with self.assertRaises(IntegrityError):
            tracking_recorder.insert_tracking(tracking=tracking1)
        with self.assertRaises(IntegrityError):
            tracking_recorder.insert_tracking(tracking=tracking2)
        with self.assertRaises(IntegrityError):
            tracking_recorder.insert_tracking(tracking=tracking3)

        # Get max tracking ID.
        self.assertEqual(tracking_recorder.max_tracking_id("upstream1"), 22)
        self.assertEqual(tracking_recorder.max_tracking_id("upstream2"), 21)
        self.assertIsNone(tracking_recorder.max_tracking_id("upstream3"))

        # Check if an event notification has been processed.
        self.assertTrue(tracking_recorder.has_tracking_id("upstream1", None))
        self.assertTrue(tracking_recorder.has_tracking_id("upstream1", 20))
        self.assertTrue(tracking_recorder.has_tracking_id("upstream1", 21))
        self.assertTrue(tracking_recorder.has_tracking_id("upstream1", 22))
        self.assertFalse(tracking_recorder.has_tracking_id("upstream1", 23))

        self.assertTrue(tracking_recorder.has_tracking_id("upstream2", None))
        self.assertTrue(tracking_recorder.has_tracking_id("upstream2", 20))
        self.assertTrue(tracking_recorder.has_tracking_id("upstream2", 21))
        self.assertFalse(tracking_recorder.has_tracking_id("upstream2", 22))

        self.assertTrue(tracking_recorder.has_tracking_id("upstream2", None))
        self.assertFalse(tracking_recorder.has_tracking_id("upstream2", 22))

        # Construct more tracking objects.
        tracking4 = Tracking("upstream1", 23)
        tracking5 = Tracking("upstream1", 24)

        tracking_recorder.insert_tracking(tracking5)

        # Can't fill in the gap.
        with self.assertRaises(IntegrityError):
            tracking_recorder.insert_tracking(tracking4)

        self.assertTrue(tracking_recorder.has_tracking_id("upstream1", 23))

    def test_wait(self) -> None:
        tracking_recorder = self.create_recorder()

        tracking_recorder.wait("upstream1", None)

        with self.assertRaises(TimeoutError):
            tracking_recorder.wait("upstream1", 21, timeout=0.1)

        tracking1 = Tracking(notification_id=21, application_name="upstream1")
        tracking_recorder.insert_tracking(tracking=tracking1)
        tracking_recorder.wait("upstream1", None)
        tracking_recorder.wait("upstream1", 10)
        tracking_recorder.wait("upstream1", 21)
        with self.assertRaises(TimeoutError):
            tracking_recorder.wait("upstream1", 22, timeout=0.1)
        with self.assertRaises(WaitInterruptedError):
            interrupt = Event()
            interrupt.set()
            tracking_recorder.wait("upstream1", 22, interrupt=interrupt)


class ProcessRecorderTestCase(RecorderTestCase, ABC):
    @abstractmethod
    def create_recorder(self) -> ProcessRecorder:
        """"""

    def test_insert_select(self) -> None:
        # Construct the recorder.
        recorder = self.create_recorder()

        # Get current position.
        self.assertIsNone(recorder.max_tracking_id("upstream_app"))

        # Write two stored events.
        originator_id1 = self.new_originator_id()
        originator_id2 = self.new_originator_id()

        stored_event1 = StoredEvent(
            originator_id=originator_id1,
            originator_version=1,
            topic="topic1",
            state=b"state1",
        )
        stored_event2 = StoredEvent(
            originator_id=originator_id1,
            originator_version=2,
            topic="topic2",
            state=b"state2",
        )
        stored_event3 = StoredEvent(
            originator_id=originator_id2,
            originator_version=1,
            topic="topic3",
            state=b"state3",
        )
        stored_event4 = StoredEvent(
            originator_id=originator_id2,
            originator_version=2,
            topic="topic4",
            state=b"state4",
        )
        tracking1 = Tracking(
            application_name="upstream_app",
            notification_id=1,
        )
        tracking2 = Tracking(
            application_name="upstream_app",
            notification_id=2,
        )

        # Insert two events with tracking info.
        recorder.insert_events(
            stored_events=[
                stored_event1,
                stored_event2,
            ],
            tracking=tracking1,
        )

        # Check get record conflict error if attempt to store same event again.
        with self.assertRaises(IntegrityError):
            recorder.insert_events(
                stored_events=[stored_event2],
                tracking=tracking2,
            )

        # Get current position.
        self.assertEqual(
            recorder.max_tracking_id("upstream_app"),
            1,
        )

        # Check can't insert third event with same tracking info.
        with self.assertRaises(IntegrityError):
            recorder.insert_events(
                stored_events=[stored_event3],
                tracking=tracking1,
            )

        # Get current position.
        self.assertEqual(
            recorder.max_tracking_id("upstream_app"),
            1,
        )

        # Insert third event with different tracking info.
        recorder.insert_events(
            stored_events=[stored_event3],
            tracking=tracking2,
        )

        # Get current position.
        self.assertEqual(
            recorder.max_tracking_id("upstream_app"),
            2,
        )

        # Insert fourth event without tracking info.
        recorder.insert_events(
            stored_events=[stored_event4],
        )

        # Get current position.
        self.assertEqual(
            recorder.max_tracking_id("upstream_app"),
            2,
        )

    def test_has_tracking_id(self) -> None:
        # Construct the recorder.
        recorder = self.create_recorder()

        self.assertTrue(recorder.has_tracking_id("upstream_app", None))
        self.assertFalse(recorder.has_tracking_id("upstream_app", 1))
        self.assertFalse(recorder.has_tracking_id("upstream_app", 2))
        self.assertFalse(recorder.has_tracking_id("upstream_app", 3))
        self.assertFalse(recorder.has_tracking_id("upstream_app", 4))

        tracking1 = Tracking(
            application_name="upstream_app",
            notification_id=1,
        )
        tracking3 = Tracking(
            application_name="upstream_app",
            notification_id=3,
        )

        recorder.insert_events(
            stored_events=[],
            tracking=tracking1,
        )

        self.assertTrue(recorder.has_tracking_id("upstream_app", 1))
        self.assertFalse(recorder.has_tracking_id("upstream_app", 2))
        self.assertFalse(recorder.has_tracking_id("upstream_app", 3))
        self.assertFalse(recorder.has_tracking_id("upstream_app", 4))

        recorder.insert_events(
            stored_events=[],
            tracking=tracking3,
        )

        self.assertTrue(recorder.has_tracking_id("upstream_app", 1))
        self.assertTrue(recorder.has_tracking_id("upstream_app", 2))
        self.assertTrue(recorder.has_tracking_id("upstream_app", 3))
        self.assertFalse(recorder.has_tracking_id("upstream_app", 4))

    def test_raises_when_lower_inserted_later(self) -> None:
        # Construct the recorder.
        recorder = self.create_recorder()

        tracking1 = Tracking(
            application_name="upstream_app",
            notification_id=1,
        )
        tracking2 = Tracking(
            application_name="upstream_app",
            notification_id=2,
        )

        # Insert tracking info.
        recorder.insert_events(
            stored_events=[],
            tracking=tracking2,
        )

        # Get current position.
        self.assertEqual(
            recorder.max_tracking_id("upstream_app"),
            2,
        )

        # Insert tracking info.
        with self.assertRaises(IntegrityError):
            recorder.insert_events(
                stored_events=[],
                tracking=tracking1,
            )

        # Get current position.
        self.assertEqual(
            recorder.max_tracking_id("upstream_app"),
            2,
        )

    def test_performance(self) -> None:
        # Construct the recorder.
        recorder = self.create_recorder()

        number = 100

        notification_ids = iter(range(1, number + 1))

        def insert_events() -> None:
            originator_id = self.new_originator_id()

            stored_event = StoredEvent(
                originator_id=originator_id,
                originator_version=0,
                topic="topic1",
                state=b"state1",
            )
            tracking1 = Tracking(
                application_name="upstream_app",
                notification_id=next(notification_ids),
            )

            recorder.insert_events(
                stored_events=[
                    stored_event,
                ],
                tracking=tracking1,
            )

        duration = timeit(insert_events, number=number)
        print(
            f"\n{self}",
            f"{1000000 * duration / number:.1f} μs per insert, "
            f"{number / duration:.0f} inserts per second",
        )


class NonInterleavingNotificationIDsBaseCase(RecorderTestCase, ABC):
    insert_num = 1000

    def test(self) -> None:
        recorder = self.create_recorder()

        max_notification_id = recorder.max_notification_id()

        race_started = Event()

        originator1_id = self.new_originator_id()
        originator2_id = self.new_originator_id()

        stack1 = self.create_stack(originator1_id)
        stack2 = self.create_stack(originator2_id)

        errors = []

        def insert_stack(stack: Sequence[StoredEvent]) -> None:
            try:
                race_started.wait()
                recorder.insert_events(stack)
            except Exception as e:
                errors.append(e)

        thread1 = Thread(target=insert_stack, args=(stack1,), daemon=True)
        thread2 = Thread(target=insert_stack, args=(stack2,), daemon=True)

        thread1.start()
        thread2.start()

        race_started.set()

        thread1.join()
        thread2.join()

        if errors:
            raise errors[0]

        # sleep(1)  # Added to make eventsourcing-axon tests work.
        notifications = recorder.select_notifications(
            start=max_notification_id,
            limit=2 * self.insert_num,
            inclusive_of_start=False,
        )
        ids_for_sequence1 = [
            e.id for e in notifications if e.originator_id == originator1_id
        ]
        ids_for_sequence2 = [
            e.id for e in notifications if e.originator_id == originator2_id
        ]
        self.assertEqual(self.insert_num, len(ids_for_sequence1))
        self.assertEqual(self.insert_num, len(ids_for_sequence2))

        max_id_for_sequence1 = max(ids_for_sequence1)
        max_id_for_sequence2 = max(ids_for_sequence2)
        min_id_for_sequence1 = min(ids_for_sequence1)
        min_id_for_sequence2 = min(ids_for_sequence2)

        if max_id_for_sequence1 > min_id_for_sequence2:
            self.assertGreater(min_id_for_sequence1, max_id_for_sequence2)
        else:
            self.assertGreater(min_id_for_sequence2, max_id_for_sequence1)

    def create_stack(self, originator_id: UUID | str) -> Sequence[StoredEvent]:
        return [
            StoredEvent(
                originator_id=originator_id,
                originator_version=i,
                topic="non-interleaving-test-event",
                state=b"{}",
            )
            for i in range(self.insert_num)
        ]

    @abstractmethod
    def create_recorder(self) -> ApplicationRecorder:
        pass


_TInfrastrutureFactory = TypeVar(
    "_TInfrastrutureFactory", bound=InfrastructureFactory[Any]
)


class InfrastructureFactoryTestCase(ABC, TestCase, Generic[_TInfrastrutureFactory]):
    env: Environment

    @abstractmethod
    def expected_factory_class(self) -> type[_TInfrastrutureFactory]:
        pass

    @abstractmethod
    def expected_aggregate_recorder_class(self) -> type[AggregateRecorder]:
        pass

    @abstractmethod
    def expected_application_recorder_class(self) -> type[ApplicationRecorder]:
        pass

    @abstractmethod
    def expected_tracking_recorder_class(self) -> type[TrackingRecorder]:
        pass

    @abstractmethod
    def tracking_recorder_subclass(self) -> type[TrackingRecorder]:
        pass

    @abstractmethod
    def expected_process_recorder_class(self) -> type[ProcessRecorder]:
        pass

    def setUp(self) -> None:
        self.factory = cast(
            _TInfrastrutureFactory, InfrastructureFactory.construct(self.env)
        )
        self.assertIsInstance(self.factory, self.expected_factory_class())
        self.transcoder = JSONTranscoder()
        self.transcoder.register(UUIDAsHex())
        self.transcoder.register(DecimalAsStr())
        self.transcoder.register(DatetimeAsISO())

    def tearDown(self) -> None:
        self.factory.close()

    def test_mapper(self) -> None:
        # Want to construct:
        #  - application recorder
        #  - snapshot recorder
        #  - mapper
        #  - event store
        #  - snapshot store

        # Want to make configurable:
        #  - cipher (and cipher key)
        #  - compressor
        #  - application recorder class (and db uri, and session)
        #  - snapshot recorder class (and db uri, and session)

        # Common environment:
        #  - factory topic
        #  - cipher topic
        #  - cipher key
        #  - compressor topic

        # POPO environment:

        # SQLite environment:
        #  - database topic
        #  - table name for stored events
        #  - table name for snapshots

        # Create mapper.

        mapper: Mapper[UUID] = self.factory.mapper(
            transcoder=self.transcoder,
        )
        self.assertIsInstance(mapper, Mapper)
        self.assertIsNone(mapper.cipher)
        self.assertIsNone(mapper.compressor)

    def test_mapper_with_compressor(self) -> None:
        # Create mapper with compressor class as topic.
        self.env[self.factory.COMPRESSOR_TOPIC] = get_topic(ZlibCompressor)
        mapper: Mapper[UUID] = self.factory.mapper(transcoder=self.transcoder)
        self.assertIsInstance(mapper, Mapper)
        self.assertIsInstance(mapper.compressor, ZlibCompressor)
        self.assertIsNone(mapper.cipher)

        # Create mapper with compressor module as topic.
        self.env[self.factory.COMPRESSOR_TOPIC] = "zlib"
        mapper = self.factory.mapper(transcoder=self.transcoder)
        self.assertIsInstance(mapper, Mapper)
        self.assertEqual(mapper.compressor, zlib)
        self.assertIsNone(mapper.cipher)

    def test_mapper_with_cipher(self) -> None:
        # Check cipher needs a key.
        self.env[self.factory.CIPHER_TOPIC] = get_topic(AESCipher)

        with self.assertRaises(EnvironmentError):
            self.factory.mapper(transcoder=self.transcoder)

        # Check setting key but no topic defers to AES.
        del self.env[self.factory.CIPHER_TOPIC]

        cipher_key = AESCipher.create_key(16)
        self.env[AESCipher.CIPHER_KEY] = cipher_key

        # Create mapper with cipher.
        mapper: Mapper[UUID] = self.factory.mapper(transcoder=self.transcoder)
        self.assertIsInstance(mapper, Mapper)
        self.assertIsNotNone(mapper.cipher)
        self.assertIsNone(mapper.compressor)

    def test_mapper_with_cipher_and_compressor(
        self,
    ) -> None:
        # Create mapper with cipher and compressor.
        self.env[self.factory.COMPRESSOR_TOPIC] = get_topic(ZlibCompressor)

        self.env[self.factory.CIPHER_TOPIC] = get_topic(AESCipher)
        cipher_key = AESCipher.create_key(16)
        self.env[AESCipher.CIPHER_KEY] = cipher_key

        mapper: Mapper[UUID] = self.factory.mapper(transcoder=self.transcoder)
        self.assertIsInstance(mapper, Mapper)
        self.assertIsNotNone(mapper.cipher)
        self.assertIsNotNone(mapper.compressor)

    def test_mapper_with_wrong_cipher_key(self) -> None:
        self.env.name = "App1"
        self.env[self.factory.CIPHER_TOPIC] = get_topic(AESCipher)
        cipher_key1 = AESCipher.create_key(16)
        cipher_key2 = AESCipher.create_key(16)
        self.env["APP1_" + AESCipher.CIPHER_KEY] = cipher_key1
        self.env["APP2_" + AESCipher.CIPHER_KEY] = cipher_key2

        mapper1: Mapper[UUID] = self.factory.mapper(
            transcoder=self.transcoder,
        )

        domain_event = DomainEvent(
            originator_id=uuid4(),
            originator_version=1,
            timestamp=DomainEvent.create_timestamp(),
        )
        stored_event = mapper1.to_stored_event(domain_event)
        copy = mapper1.to_domain_event(stored_event)
        self.assertEqual(domain_event.originator_id, copy.originator_id)

        self.env.name = "App2"
        mapper2: Mapper[UUID] = self.factory.mapper(
            transcoder=self.transcoder,
        )
        # This should fail because the infrastructure factory
        # should read different cipher keys from the environment.
        with self.assertRaises(ValueError):
            mapper2.to_domain_event(stored_event)

    def test_create_aggregate_recorder(self) -> None:
        recorder = self.factory.aggregate_recorder()
        self.assertEqual(type(recorder), self.expected_aggregate_recorder_class())

        self.assertIsInstance(recorder, AggregateRecorder)

        # Exercise code path where table is not created.
        self.env["CREATE_TABLE"] = "f"
        recorder = self.factory.aggregate_recorder()
        self.assertEqual(type(recorder), self.expected_aggregate_recorder_class())

    def test_create_application_recorder(self) -> None:
        recorder = self.factory.application_recorder()
        self.assertEqual(type(recorder), self.expected_application_recorder_class())
        self.assertIsInstance(recorder, ApplicationRecorder)

        # Exercise code path where table is not created.
        self.env["CREATE_TABLE"] = "f"
        recorder = self.factory.application_recorder()
        self.assertEqual(type(recorder), self.expected_application_recorder_class())

    def test_create_tracking_recorder(self) -> None:
        recorder = self.factory.tracking_recorder()
        self.assertEqual(type(recorder), self.expected_tracking_recorder_class())
        self.assertIsInstance(recorder, TrackingRecorder)

        # Exercise code path where table is not created.
        self.env["CREATE_TABLE"] = "f"
        recorder = self.factory.tracking_recorder()
        self.assertEqual(type(recorder), self.expected_tracking_recorder_class())

        # Exercise code path where tracking recorder class is specified as arg.
        subclass = self.tracking_recorder_subclass()
        recorder = self.factory.tracking_recorder(subclass)
        self.assertEqual(type(recorder), subclass)

        # Exercise code path where tracking recorder class is specified as topic.
        self.factory.env[self.factory.TRACKING_RECORDER_TOPIC] = get_topic(subclass)
        recorder = self.factory.tracking_recorder()
        self.assertEqual(type(recorder), subclass)

    def test_create_process_recorder(self) -> None:
        recorder = self.factory.process_recorder()
        self.assertEqual(type(recorder), self.expected_process_recorder_class())
        self.assertIsInstance(recorder, ProcessRecorder)

        # Exercise code path where table is not created.
        self.env["CREATE_TABLE"] = "f"
        recorder = self.factory.process_recorder()
        self.assertEqual(type(recorder), self.expected_process_recorder_class())


def tmpfile_uris() -> Iterator[str]:
    tmp_files = []
    ram_disk_path = Path("/Volumes/RAM DISK/")
    prefix: str | None = None
    if ram_disk_path.exists():
        prefix = str(ram_disk_path)
    while True:
        with NamedTemporaryFile(
            prefix=prefix,
            suffix="_eventsourcing_test.db",
        ) as tmp_file:
            tmp_files.append(tmp_file)
            yield "file:" + tmp_file.name


class CustomType1:
    def __init__(self, value: UUID):
        self.value = value

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and self.__dict__ == other.__dict__

    def __hash__(self) -> int:
        raise NotImplementedError


class CustomType2:
    def __init__(self, value: CustomType1):
        self.value = value

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and self.__dict__ == other.__dict__

    def __hash__(self) -> int:
        raise NotImplementedError


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class Mydict(dict[_KT, _VT]):  # noqa: PLW1641
    def __repr__(self) -> str:
        return f"{type(self).__name__}({super().__repr__()})"

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and super().__eq__(other)


_T = TypeVar("_T")


class MyList(list[_T]):  # noqa: PLW1641
    def __repr__(self) -> str:
        return f"{type(self).__name__}({super().__repr__()})"

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and super().__eq__(other)


class MyStr(str):
    __slots__ = ()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({super().__repr__()})"

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and super().__eq__(other)

    def __hash__(self) -> int:
        return hash(str(self))


class MyInt(int):
    def __repr__(self) -> str:
        return f"{type(self).__name__}({super().__repr__()})"

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and super().__eq__(other)

    def __hash__(self) -> int:
        return int(self)


class MyClass:
    pass


class CustomType1AsDict(Transcoding):
    type = CustomType1
    name = "custom_type1_as_dict"

    def encode(self, obj: CustomType1) -> UUID:
        return obj.value

    def decode(self, data: UUID) -> CustomType1:
        assert isinstance(data, UUID)
        return CustomType1(value=data)


class CustomType2AsDict(Transcoding):
    type = CustomType2
    name = "custom_type2_as_dict"

    def encode(self, obj: CustomType2) -> CustomType1:
        return obj.value

    def decode(self, data: CustomType1) -> CustomType2:
        assert isinstance(data, CustomType1)
        return CustomType2(data)


class TranscoderTestCase(TestCase):
    def setUp(self) -> None:
        self.transcoder = self.construct_transcoder()

    def construct_transcoder(self) -> Transcoder:
        raise NotImplementedError

    def test_str(self) -> None:
        obj = "a"
        data = self.transcoder.encode(obj)
        self.assertEqual(data, b'"a"')
        self.assertEqual(obj, self.transcoder.decode(data))

        obj = "abc"
        data = self.transcoder.encode(obj)
        self.assertEqual(data, b'"abc"')
        self.assertEqual(obj, self.transcoder.decode(data))

        obj = "a'b"
        data = self.transcoder.encode(obj)
        self.assertEqual(data, b'''"a'b"''')
        self.assertEqual(obj, self.transcoder.decode(data))

        obj = 'a"b'
        data = self.transcoder.encode(obj)
        self.assertEqual(data, b'''"a\\"b"''')
        self.assertEqual(obj, self.transcoder.decode(data))

        obj = "🐈 哈哈"
        data = self.transcoder.encode(obj)
        self.assertEqual(b'"\xf0\x9f\x90\x88 \xe5\x93\x88\xe5\x93\x88"', data)
        self.assertEqual(obj, self.transcoder.decode(data))

        # Check data encoded with ensure_ascii=True can be decoded okay.
        legacy_encoding_with_ensure_ascii_true = b'"\\ud83d\\udc08 \\u54c8\\u54c8"'
        self.assertEqual(
            obj, self.transcoder.decode(legacy_encoding_with_ensure_ascii_true)
        )

    def test_dict(self) -> None:
        # Empty dict.
        obj1: dict[Never, Never] = {}
        data = self.transcoder.encode(obj1)
        self.assertEqual(data, b"{}")
        self.assertEqual(obj1, self.transcoder.decode(data))

        # dict with single key.
        obj2 = {"a": 1}
        data = self.transcoder.encode(obj2)
        self.assertEqual(data, b'{"a":1}')
        self.assertEqual(obj2, self.transcoder.decode(data))

        # dict with many keys.
        obj3 = {"a": 1, "b": 2}
        data = self.transcoder.encode(obj3)
        self.assertEqual(data, b'{"a":1,"b":2}')
        self.assertEqual(obj3, self.transcoder.decode(data))

        # Empty dict in dict.
        obj4: dict[str, dict[Never, Never]] = {"a": {}}
        data = self.transcoder.encode(obj4)
        self.assertEqual(data, b'{"a":{}}')
        self.assertEqual(obj4, self.transcoder.decode(data))

        # Empty dicts in dict.
        obj5: dict[str, dict[Never, Never]] = {"a": {}, "b": {}}
        data = self.transcoder.encode(obj5)
        self.assertEqual(data, b'{"a":{},"b":{}}')
        self.assertEqual(obj5, self.transcoder.decode(data))

        # Empty dict in dict in dict.
        obj6: dict[str, dict[str, dict[Never, Never]]] = {"a": {"b": {}}}
        data = self.transcoder.encode(obj6)
        self.assertEqual(data, b'{"a":{"b":{}}}')
        self.assertEqual(obj6, self.transcoder.decode(data))

        # Int in dict in dict in dict.
        obj7 = {"a": {"b": {"c": 1}}}
        data = self.transcoder.encode(obj7)
        self.assertEqual(data, b'{"a":{"b":{"c":1}}}')
        self.assertEqual(obj7, self.transcoder.decode(data))

        # TODO: Int keys?
        # obj = {1: "a"}
        # data = self.transcoder.encode(obj)
        # self.assertEqual(data, b'{1:{"a"}')
        # self.assertEqual(obj, self.transcoder.decode(data))

    def test_dict_with_len_2_and__data_(self) -> None:
        obj = {"_data_": 1, "something_else": 2}
        data = self.transcoder.encode(obj)
        self.assertEqual(obj, self.transcoder.decode(data))

    def test_dict_with_len_2_and__type_(self) -> None:
        obj = {"_type_": 1, "something_else": 2}
        data = self.transcoder.encode(obj)
        self.assertEqual(obj, self.transcoder.decode(data))

    def test_dict_subclass(self) -> None:
        my_dict = Mydict({"a": 1})
        data = self.transcoder.encode(my_dict)
        self.assertEqual(b'{"_type_":"mydict","_data_":{"a":1}}', data)
        copy = self.transcoder.decode(data)
        self.assertEqual(my_dict, copy)

    def test_list_subclass(self) -> None:
        my_list = MyList((("a", 1),))
        data = self.transcoder.encode(my_list)
        copy = self.transcoder.decode(data)
        self.assertEqual(my_list, copy)

    def test_str_subclass(self) -> None:
        my_str = MyStr("a")
        data = self.transcoder.encode(my_str)
        copy = self.transcoder.decode(data)
        self.assertEqual(my_str, copy)

    def test_int_subclass(self) -> None:
        my_int = MyInt(3)
        data = self.transcoder.encode(my_int)
        copy = self.transcoder.decode(data)
        self.assertEqual(my_int, copy)

    def test_tuple(self) -> None:
        # Empty tuple.
        obj1 = ()
        data = self.transcoder.encode(obj1)
        self.assertEqual(data, b'{"_type_":"tuple_as_list","_data_":[]}')
        self.assertEqual(obj1, self.transcoder.decode(data))

        # Empty tuple in a tuple.
        obj2 = ((),)
        data = self.transcoder.encode(obj2)
        self.assertEqual(obj2, self.transcoder.decode(data))

        # Int in tuple in a tuple.
        obj3 = ((1, 2),)
        data = self.transcoder.encode(obj3)
        self.assertEqual(obj3, self.transcoder.decode(data))

        # Str in tuple in a tuple.
        obj4 = (("a", "b"),)
        data = self.transcoder.encode(obj4)
        self.assertEqual(obj4, self.transcoder.decode(data))

        # Int and str in tuple in a tuple.
        obj5 = ((1, "a"),)
        data = self.transcoder.encode(obj5)
        self.assertEqual(obj5, self.transcoder.decode(data))

    def test_list(self) -> None:
        # Empty list.
        obj1: list[Never] = []
        data = self.transcoder.encode(obj1)
        self.assertEqual(obj1, self.transcoder.decode(data))

        # Empty list in a list.
        obj2: list[list[Never]] = [[]]
        data = self.transcoder.encode(obj2)
        self.assertEqual(obj2, self.transcoder.decode(data))

        # Int in list in a list.
        obj3 = [[1, 2]]
        data = self.transcoder.encode(obj3)
        self.assertEqual(obj3, self.transcoder.decode(data))

        # Str in list in a list.
        obj4 = [["a", "b"]]
        data = self.transcoder.encode(obj4)
        self.assertEqual(obj4, self.transcoder.decode(data))

        # Int and str in list in a list.
        obj5 = [[1, "a"]]
        data = self.transcoder.encode(obj5)
        self.assertEqual(obj5, self.transcoder.decode(data))

    def test_mixed(self) -> None:
        obj1 = [(1, "a"), {"b": 2}]
        data = self.transcoder.encode(obj1)
        self.assertEqual(obj1, self.transcoder.decode(data))

        obj2 = ([1, "a"], {"b": 2})
        data = self.transcoder.encode(obj2)
        self.assertEqual(obj2, self.transcoder.decode(data))

        obj3 = {"a": (1, 2), "b": [3, 4]}
        data = self.transcoder.encode(obj3)
        self.assertEqual(obj3, self.transcoder.decode(data))

    def test_custom_type_in_dict(self) -> None:
        # Int in dict in dict in dict.
        obj = {"a": CustomType2(CustomType1(UUID("b2723fe2c01a40d2875ea3aac6a09ff5")))}
        data = self.transcoder.encode(obj)
        decoded_obj = self.transcoder.decode(data)
        self.assertEqual(obj, decoded_obj)

    def test_nested_custom_type(self) -> None:
        obj = CustomType2(CustomType1(UUID("b2723fe2c01a40d2875ea3aac6a09ff5")))
        data = self.transcoder.encode(obj)
        expect = (
            b'{"_type_":"custom_type2_as_dict","_data_":'
            b'{"_type_":"custom_type1_as_dict","_data_":'
            b'{"_type_":"uuid_hex","_data_":"b2723fe2c01'
            b'a40d2875ea3aac6a09ff5"}}}'
        )
        self.assertEqual(data, expect)
        copy = self.transcoder.decode(data)
        self.assertIsInstance(copy, CustomType2)
        self.assertIsInstance(copy.value, CustomType1)
        self.assertIsInstance(copy.value.value, UUID)
        self.assertEqual(copy.value.value, obj.value.value)

    def test_custom_type_error(self) -> None:
        # Expect a TypeError when encoding because transcoding not registered.
        with self.assertRaises(TypeError) as cm:
            self.transcoder.encode(MyClass())

        self.assertEqual(
            cm.exception.args[0],
            "Object of type <class 'eventsourcing.tests.persistence."
            "MyClass'> is not serializable. Please define "
            "and register a custom transcoding for this type.",
        )

        # Expect a TypeError when encoding because transcoding not registered (nested).
        with self.assertRaises(TypeError) as cm:
            self.transcoder.encode({"a": MyClass()})

        self.assertEqual(
            cm.exception.args[0],
            "Object of type <class 'eventsourcing.tests.persistence."
            "MyClass'> is not serializable. Please define "
            "and register a custom transcoding for this type.",
        )

        # Check we get a TypeError when decoding because transcodings aren't registered.
        data = b'{"_type_":"custom_type3_as_dict","_data_":""}'

        with self.assertRaises(TypeError) as cm:
            self.transcoder.decode(data)

        self.assertEqual(
            cm.exception.args[0],
            "Data serialized with name 'custom_type3_as_dict' is not "
            "deserializable. Please register a custom transcoding for this type.",
        )



from __future__ import annotations

import sqlite3
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Literal, cast
from uuid import UUID

from eventsourcing.persistence import (
    AggregateRecorder,
    ApplicationRecorder,
    Connection,
    ConnectionPool,
    Cursor,
    DatabaseError,
    DataError,
    InfrastructureFactory,
    IntegrityError,
    InterfaceError,
    InternalError,
    Notification,
    NotSupportedError,
    OperationalError,
    PersistenceError,
    ProcessRecorder,
    ProgrammingError,
    Recorder,
    StoredEvent,
    Subscription,
    Tracking,
    TrackingRecorder,
)
from eventsourcing.utils import Environment, EnvType, resolve_topic, strtobool

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from types import TracebackType

SQLITE3_DEFAULT_LOCK_TIMEOUT = 5


class SQLiteCursor(Cursor):
    def __init__(self, sqlite_cursor: sqlite3.Cursor):
        self.sqlite_cursor = sqlite_cursor

    def __enter__(self) -> sqlite3.Cursor:
        return self.sqlite_cursor

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        self.sqlite_cursor.close()

    def execute(self, *args: Any, **kwargs: Any) -> None:
        self.sqlite_cursor.execute(*args, **kwargs)

    def executemany(self, *args: Any, **kwargs: Any) -> None:
        self.sqlite_cursor.executemany(*args, **kwargs)

    def fetchall(self) -> Any:
        return self.sqlite_cursor.fetchall()

    def fetchone(self) -> Any:
        return self.sqlite_cursor.fetchone()

    @property
    def lastrowid(self) -> Any:
        return self.sqlite_cursor.lastrowid


class SQLiteConnection(Connection[SQLiteCursor]):
    def __init__(self, sqlite_conn: sqlite3.Connection, max_age: float | None):
        super().__init__(max_age=max_age)
        self._sqlite_conn = sqlite_conn

    @contextmanager
    def transaction(self, *, commit: bool) -> Iterator[SQLiteCursor]:
        # Context managed cursor, and context managed transaction.
        with SQLiteTransaction(self, commit=commit) as curs, curs:
            yield curs

    def cursor(self) -> SQLiteCursor:
        return SQLiteCursor(self._sqlite_conn.cursor())

    def rollback(self) -> None:
        self._sqlite_conn.rollback()

    def commit(self) -> None:
        self._sqlite_conn.commit()

    def _close(self) -> None:
        self._sqlite_conn.close()
        super()._close()


class SQLiteTransaction:
    def __init__(self, connection: SQLiteConnection, *, commit: bool = False):
        self.connection = connection
        self.commit = commit

    def __enter__(self) -> SQLiteCursor:
        # We must issue a "BEGIN" explicitly
        # when running in auto-commit mode.
        cursor = self.connection.cursor()
        cursor.execute("BEGIN")
        return cursor

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        try:
            if exc_val:
                # Roll back all changes
                # if an exception occurs.
                self.connection.rollback()
                raise exc_val
            if not self.commit:
                self.connection.rollback()
            else:
                self.connection.commit()
        except sqlite3.InterfaceError as e:
            raise InterfaceError(e) from e
        except sqlite3.DataError as e:
            raise DataError(e) from e
        except sqlite3.OperationalError as e:
            raise OperationalError(e) from e
        except sqlite3.IntegrityError as e:
            raise IntegrityError(e) from e
        except sqlite3.InternalError as e:
            raise InternalError(e) from e
        except sqlite3.ProgrammingError as e:
            raise ProgrammingError(e) from e
        except sqlite3.NotSupportedError as e:
            raise NotSupportedError(e) from e
        except sqlite3.DatabaseError as e:
            raise DatabaseError(e) from e
        except sqlite3.Error as e:
            raise PersistenceError(e) from e


class SQLiteConnectionPool(ConnectionPool[SQLiteConnection]):
    def __init__(
        self,
        *,
        db_name: str,
        lock_timeout: int | None = None,
        pool_size: int = 5,
        max_overflow: int = 10,
        pool_timeout: float = 5.0,
        max_age: float | None = None,
        pre_ping: bool = False,
    ):
        self.db_name = db_name
        self.lock_timeout = lock_timeout
        self.is_sqlite_memory_mode = self.detect_memory_mode(db_name)
        self.is_journal_mode_wal = False
        self.journal_mode_was_changed_to_wal = False
        super().__init__(
            pool_size=pool_size,
            max_overflow=max_overflow,
            pool_timeout=pool_timeout,
            max_age=max_age,
            pre_ping=pre_ping,
            mutually_exclusive_read_write=self.is_sqlite_memory_mode,
        )

    @staticmethod
    def detect_memory_mode(db_name: str) -> bool:
        return bool(db_name) and (":memory:" in db_name or "mode=memory" in db_name)

    def _create_connection(self) -> SQLiteConnection:
        # Make a connection to an SQLite database.
        try:
            c = sqlite3.connect(
                database=self.db_name,
                uri=True,
                check_same_thread=False,
                isolation_level=None,  # Auto-commit mode.
                cached_statements=True,
                timeout=self.lock_timeout or SQLITE3_DEFAULT_LOCK_TIMEOUT,
            )
        except (sqlite3.Error, TypeError) as e:
            raise InterfaceError(e) from e

        # Use WAL (write-ahead log) mode if file-based database.
        if not self.is_sqlite_memory_mode and not self.is_journal_mode_wal:
            cursor = c.cursor()
            cursor.execute("PRAGMA journal_mode;")
            mode = cursor.fetchone()[0]
            if mode.lower() == "wal":
                self.is_journal_mode_wal = True
            else:
                cursor.execute("PRAGMA journal_mode=WAL;")
                self.is_journal_mode_wal = True
                self.journal_mode_was_changed_to_wal = True

        # Set the row factory.
        c.row_factory = sqlite3.Row

        # Return the connection.
        return SQLiteConnection(sqlite_conn=c, max_age=self.max_age)


class SQLiteDatastore:
    def __init__(
        self,
        db_name: str,
        *,
        lock_timeout: int | None = None,
        pool_size: int = 5,
        max_overflow: int = 10,
        pool_timeout: float = 5.0,
        max_age: float | None = None,
        pre_ping: bool = False,
        single_row_tracking: bool = True,
        originator_id_type: Literal["uuid", "text"] = "uuid",
    ):
        self.pool = SQLiteConnectionPool(
            db_name=db_name,
            lock_timeout=lock_timeout,
            pool_size=pool_size,
            max_overflow=max_overflow,
            pool_timeout=pool_timeout,
            max_age=max_age,
            pre_ping=pre_ping,
        )
        self.single_row_tracking = single_row_tracking
        self.originator_id_type = originator_id_type

    @contextmanager
    def transaction(self, *, commit: bool) -> Iterator[SQLiteCursor]:
        connection = self.get_connection(commit=commit)
        with connection as conn, conn.transaction(commit=commit) as curs:
            yield curs

    @contextmanager
    def get_connection(self, *, commit: bool) -> Iterator[SQLiteConnection]:
        # Using reader-writer interlocking is necessary for in-memory databases,
        # but also speeds up (and provides "fairness") to file-based databases.
        conn = self.pool.get_connection(is_writer=commit)
        try:
            yield conn
        finally:
            self.pool.put_connection(conn)

    def close(self) -> None:
        self.pool.close()

    def __del__(self) -> None:
        self.close()


class SQLiteRecorder(Recorder):
    def __init__(
        self,
        datastore: SQLiteDatastore,
    ):
        assert isinstance(datastore, SQLiteDatastore)
        self.datastore = datastore
        self.create_table_statements = self.construct_create_table_statements()

    def construct_create_table_statements(self) -> list[str]:
        return []

    def create_table(self) -> None:
        with self.datastore.transaction(commit=True) as c:
            self._create_table(c)

    def _create_table(self, c: SQLiteCursor) -> None:
        for statement in self.create_table_statements:
            c.execute(statement)

    def convert_originator_id(self, originator_id: str) -> UUID | str:
        return (
            UUID(originator_id)
            if self.datastore.originator_id_type == "uuid"
            else originator_id
        )


class SQLiteAggregateRecorder(SQLiteRecorder, AggregateRecorder):
    def __init__(
        self,
        datastore: SQLiteDatastore,
        events_table_name: str = "stored_events",
    ):
        self.events_table_name = events_table_name
        super().__init__(datastore)
        self.insert_events_statement = (
            f"INSERT INTO {self.events_table_name} VALUES (?,?,?,?)"
        )
        self.select_events_statement = (
            f"SELECT * FROM {self.events_table_name} WHERE originator_id=? "
        )

    def construct_create_table_statements(self) -> list[str]:
        statements = super().construct_create_table_statements()
        statements.append(
            "CREATE TABLE IF NOT EXISTS "
            f"{self.events_table_name} ("
            "originator_id TEXT, "
            "originator_version INTEGER, "
            "topic TEXT, "
            "state BLOB, "
            "PRIMARY KEY "
            "(originator_id, originator_version)) "
            "WITHOUT ROWID"
        )
        return statements

    def insert_events(
        self, stored_events: Sequence[StoredEvent], **kwargs: Any
    ) -> Sequence[int] | None:
        with self.datastore.transaction(commit=True) as c:
            return self._insert_events(c, stored_events, **kwargs)

    def _insert_events(
        self,
        c: SQLiteCursor,
        stored_events: Sequence[StoredEvent],
        **_: Any,
    ) -> Sequence[int] | None:
        params = [
            (
                (
                    s.originator_id.hex
                    if isinstance(s.originator_id, UUID)
                    else s.originator_id
                ),
                s.originator_version,
                s.topic,
                s.state,
            )
            for s in stored_events
        ]
        c.executemany(self.insert_events_statement, params)
        return None

    def select_events(
        self,
        originator_id: UUID | str,
        *,
        gt: int | None = None,
        lte: int | None = None,
        desc: bool = False,
        limit: int | None = None,
    ) -> Sequence[StoredEvent]:
        statement = self.select_events_statement
        params: list[Any] = [
            originator_id.hex if isinstance(originator_id, UUID) else originator_id
        ]
        if gt is not None:
            statement += "AND originator_version>? "
            params.append(gt)
        if lte is not None:
            statement += "AND originator_version<=? "
            params.append(lte)
        statement += "ORDER BY originator_version "
        if desc is False:
            statement += "ASC "
        else:
            statement += "DESC "
        if limit is not None:
            statement += "LIMIT ? "
            params.append(limit)
        with self.datastore.transaction(commit=False) as c:
            c.execute(statement, params)
            return [
                StoredEvent(
                    originator_id=self.convert_originator_id(row["originator_id"]),
                    originator_version=row["originator_version"],
                    topic=row["topic"],
                    state=row["state"],
                )
                for row in c.fetchall()
            ]


class SQLiteApplicationRecorder(
    SQLiteAggregateRecorder,
    ApplicationRecorder,
):
    def __init__(
        self,
        datastore: SQLiteDatastore,
        events_table_name: str = "stored_events",
    ):
        super().__init__(datastore, events_table_name)
        self.select_max_notification_id_statement = (
            f"SELECT MAX(rowid) FROM {self.events_table_name}"
        )

    def construct_create_table_statements(self) -> list[str]:
        statement = (
            "CREATE TABLE IF NOT EXISTS "
            f"{self.events_table_name} ("
            "originator_id TEXT, "
            "originator_version INTEGER, "
            "topic TEXT, "
            "state BLOB, "
            "PRIMARY KEY "
            "(originator_id, originator_version))"
        )
        return [statement]

    def _insert_events(
        self,
        c: SQLiteCursor,
        stored_events: Sequence[StoredEvent],
        **_: Any,
    ) -> Sequence[int] | None:
        returning = []
        for s in stored_events:
            c.execute(
                self.insert_events_statement,
                (
                    (
                        s.originator_id.hex
                        if isinstance(s.originator_id, UUID)
                        else s.originator_id
                    ),
                    s.originator_version,
                    s.topic,
                    s.state,
                ),
            )
            returning.append(c.lastrowid)
        return returning

    def select_notifications(
        self,
        start: int | None,
        limit: int,
        stop: int | None = None,
        topics: Sequence[str] = (),
        *,
        inclusive_of_start: bool = True,
    ) -> Sequence[Notification]:
        """Returns a list of event notifications
        from 'start', limited by 'limit'.
        """
        params: list[int | str] = []
        statement = f"SELECT rowid, * FROM {self.events_table_name} "
        has_where = False
        if start is not None:
            has_where = True
            statement += "WHERE "
            params.append(start)
            if inclusive_of_start:
                statement += "rowid>=? "
            else:
                statement += "rowid>? "

        if stop is not None:
            if not has_where:
                has_where = True
                statement += "WHERE "
            else:
                statement += "AND "
            params.append(stop)
            statement += "rowid<=? "

        if topics:
            if not has_where:
                statement += "WHERE "
            else:
                statement += "AND "
            params += list(topics)
            statement += f"topic IN ({','.join('?' * len(topics))}) "

        params.append(limit)
        statement += "ORDER BY rowid LIMIT ?"

        with self.datastore.transaction(commit=False) as c:
            c.execute(statement, params)
            return [
                Notification(
                    id=row["rowid"],
                    originator_id=self.convert_originator_id(row["originator_id"]),
                    originator_version=row["originator_version"],
                    topic=row["topic"],
                    state=row["state"],
                )
                for row in c.fetchall()
            ]

    def max_notification_id(self) -> int:
        """Returns the maximum notification ID."""
        with self.datastore.transaction(commit=False) as c:
            return self._max_notification_id(c)

    def _max_notification_id(self, c: SQLiteCursor) -> int:
        c.execute(self.select_max_notification_id_statement)
        return c.fetchone()[0]

    def subscribe(
        self, gt: int | None = None, topics: Sequence[str] = ()
    ) -> Subscription[ApplicationRecorder]:
        """This method is not implemented on this class."""
        msg = f"The {type(self).__qualname__} recorder does not support subscriptions"
        raise NotImplementedError(msg)


class SQLiteTrackingRecorder(SQLiteRecorder, TrackingRecorder):
    def __init__(
        self,
        datastore: SQLiteDatastore,
        **kwargs: Any,
    ):
        super().__init__(datastore, **kwargs)
        self.tracking_table_exists: bool = False
        self.tracking_migration_previous: int | None = None
        self.tracking_migration_current: int | None = None
        self.table_migration_identifier = "__migration__"
        self.has_checked_for_multi_row_tracking_table: bool = False
        if self.datastore.single_row_tracking:
            self.insert_tracking_statement = (
                "INSERT INTO tracking "
                "VALUES (:application_name, :notification_id) "
                "ON CONFLICT (application_name) DO UPDATE "
                "SET notification_id = :notification_id "
                "WHERE tracking.notification_id < :notification_id "
                "RETURNING notification_id"
            )
        else:
            self.insert_tracking_statement = (
                "INSERT INTO tracking VALUES (:application_name, :notification_id)"
            )
        self.select_max_tracking_id_statement = (
            "SELECT MAX(notification_id) FROM tracking WHERE application_name=?"
        )

    def construct_create_table_statements(self) -> list[str]:
        statements = super().construct_create_table_statements()
        if self.datastore.single_row_tracking:
            statements.append(
                "CREATE TABLE IF NOT EXISTS tracking ("
                "application_name TEXT, "
                "notification_id INTEGER, "
                "PRIMARY KEY "
                "(application_name)) "
                "WITHOUT ROWID"
            )
        else:
            statements.append(
                "CREATE TABLE IF NOT EXISTS tracking ("
                "application_name TEXT, "
                "notification_id INTEGER, "
                "PRIMARY KEY "
                "(application_name, notification_id)) "
                "WITHOUT ROWID"
            )
        return statements

    def create_table(self) -> None:
        # Get the migration version.
        try:
            self.tracking_migration_current = self.tracking_migration_previous = (
                self.max_tracking_id(self.table_migration_identifier)
            )
        except OperationalError:
            pass
        else:
            self.tracking_table_exists = True
        super().create_table()
        if (
            not self.datastore.single_row_tracking
            and self.tracking_migration_current is not None
        ):
            msg = "Can't do multi-row tracking with single-row tracking table"
            raise OperationalError(msg)

    def _create_table(self, c: SQLiteCursor) -> None:
        max_tracking_ids: dict[str, int] = {}
        if (
            self.datastore.single_row_tracking
            and self.tracking_table_exists
            and not self.tracking_migration_previous
        ):
            # Migrate tracking to use single-row per application name.
            # - Get all application names.
            c.execute("SELECT DISTINCT application_name FROM tracking")
            application_names: list[str] = [
                select_row["application_name"] for select_row in c.fetchall()
            ]

            # - Get max tracking ID for each application name.
            for application_name in application_names:
                c.execute(self.select_max_tracking_id_statement, (application_name,))
                max_tracking_id_row = c.fetchone()
                assert max_tracking_id_row is not None
                max_tracking_ids[application_name] = max_tracking_id_row[0]
            # - Rename the table.
            drop_table_statement = "ALTER TABLE tracking RENAME TO old1_tracking"
            c.execute(drop_table_statement)
        # Create the table.
        super()._create_table(c)
        # - Maybe insert migration tracking record and application tracking records.
        if self.datastore.single_row_tracking and (
            not self.tracking_table_exists
            or (self.tracking_table_exists and not self.tracking_migration_previous)
        ):
            # - Assume we just created a table for single-row tracking.
            self._insert_tracking(c, Tracking(self.table_migration_identifier, 1))
            self.tracking_migration_current = 1
            for application_name, max_tracking_id in max_tracking_ids.items():
                self._insert_tracking(c, Tracking(application_name, max_tracking_id))

    def insert_tracking(self, tracking: Tracking) -> None:
        with self.datastore.transaction(commit=True) as c:
            self._insert_tracking(c, tracking)

    def _insert_tracking(
        self,
        c: SQLiteCursor,
        tracking: Tracking,
    ) -> None:
        self._check_has_multi_row_tracking_table(c)

        c.execute(
            self.insert_tracking_statement,
            {
                "application_name": tracking.application_name,
                "notification_id": tracking.notification_id,
            },
        )
        if self.datastore.single_row_tracking:
            fetchone = c.fetchone()
            if fetchone is None:
                msg = (
                    "Failed to record tracking for "
                    f"{tracking.application_name} {tracking.notification_id}"
                )
                raise IntegrityError(msg)

    def _check_has_multi_row_tracking_table(self, c: SQLiteCursor) -> None:
        if (
            not self.datastore.single_row_tracking
            and not self.has_checked_for_multi_row_tracking_table
            and self._max_tracking_id(self.table_migration_identifier, c)
        ):
            msg = "Can't do multi-row tracking with single-row tracking table"
            raise OperationalError(msg)
        self.has_checked_for_multi_row_tracking_table = True

    def max_tracking_id(self, application_name: str) -> int | None:
        with self.datastore.transaction(commit=False) as c:
            return self._max_tracking_id(application_name, c)

    def _max_tracking_id(self, application_name: str, c: SQLiteCursor) -> int | None:
        params = [application_name]
        c.execute(self.select_max_tracking_id_statement, params)
        return c.fetchone()[0]


class SQLiteProcessRecorder(
    SQLiteTrackingRecorder,
    SQLiteApplicationRecorder,
    ProcessRecorder,
):
    def __init__(
        self,
        datastore: SQLiteDatastore,
        *,
        events_table_name: str = "stored_events",
    ):
        super().__init__(datastore, events_table_name=events_table_name)

    def _insert_events(
        self,
        c: SQLiteCursor,
        stored_events: Sequence[StoredEvent],
        **kwargs: Any,
    ) -> Sequence[int] | None:
        returning = super()._insert_events(c, stored_events, **kwargs)
        tracking: Tracking | None = kwargs.get("tracking")
        if tracking is not None:
            self._insert_tracking(c, tracking)
        return returning


class SQLiteFactory(InfrastructureFactory[SQLiteTrackingRecorder]):
    SQLITE_DBNAME = "SQLITE_DBNAME"
    SQLITE_LOCK_TIMEOUT = "SQLITE_LOCK_TIMEOUT"
    SQLITE_SINGLE_ROW_TRACKING = "SINGLE_ROW_TRACKING"
    ORIGINATOR_ID_TYPE = "ORIGINATOR_ID_TYPE"
    CREATE_TABLE = "CREATE_TABLE"

    aggregate_recorder_class = SQLiteAggregateRecorder
    application_recorder_class = SQLiteApplicationRecorder
    tracking_recorder_class = SQLiteTrackingRecorder
    process_recorder_class = SQLiteProcessRecorder

    def __init__(self, env: Environment | EnvType | None):
        super().__init__(env)
        db_name = self.env.get(self.SQLITE_DBNAME)
        if not db_name:
            msg = (
                "SQLite database name not found "
                "in environment with keys: "
                f"{', '.join(self.env.create_keys(self.SQLITE_DBNAME))}"
            )
            raise OSError(msg)

        lock_timeout_str = (
            self.env.get(self.SQLITE_LOCK_TIMEOUT) or ""
        ).strip() or None

        lock_timeout: int | None = None
        if lock_timeout_str is not None:
            try:
                lock_timeout = int(lock_timeout_str)
            except ValueError:
                msg = (
                    "SQLite environment value for key "
                    f"'{self.SQLITE_LOCK_TIMEOUT}' is invalid. "
                    "If set, an int or empty string is expected: "
                    f"'{lock_timeout_str}'"
                )
                raise OSError(msg) from None

        single_row_tracking = strtobool(
            self.env.get(self.SQLITE_SINGLE_ROW_TRACKING, "t")
        )

        originator_id_type = cast(
            Literal["uuid", "text"],
            self.env.get(self.ORIGINATOR_ID_TYPE, "uuid"),
        )
        if originator_id_type.lower() not in ("uuid", "text"):
            msg = (
                f"Invalid {self.ORIGINATOR_ID_TYPE} '{originator_id_type}', "
                f"must be 'uuid' or 'text'"
            )
            raise OSError(msg)

        self.datastore = SQLiteDatastore(
            db_name=db_name,
            lock_timeout=lock_timeout,
            single_row_tracking=single_row_tracking,
            originator_id_type=originator_id_type,
        )

    def aggregate_recorder(self, purpose: str = "events") -> AggregateRecorder:
        events_table_name = "stored_" + purpose
        recorder = self.aggregate_recorder_class(
            datastore=self.datastore,
            events_table_name=events_table_name,
        )
        if self.env_create_table():
            recorder.create_table()
        return recorder

    def application_recorder(self) -> ApplicationRecorder:
        application_recorder_topic = self.env.get(self.APPLICATION_RECORDER_TOPIC)

        if application_recorder_topic:
            application_recorder_class: type[SQLiteApplicationRecorder] = resolve_topic(
                application_recorder_topic
            )
            assert issubclass(application_recorder_class, SQLiteApplicationRecorder)
        else:
            application_recorder_class = self.application_recorder_class

        recorder = application_recorder_class(datastore=self.datastore)

        if self.env_create_table():
            recorder.create_table()
        return recorder

    def tracking_recorder(
        self, tracking_recorder_class: type[SQLiteTrackingRecorder] | None = None
    ) -> SQLiteTrackingRecorder:
        if tracking_recorder_class is None:
            tracking_recorder_topic = self.env.get(self.TRACKING_RECORDER_TOPIC)

            if tracking_recorder_topic:
                tracking_recorder_class = resolve_topic(tracking_recorder_topic)
            else:
                tracking_recorder_class = self.tracking_recorder_class

        assert tracking_recorder_class is not None
        assert issubclass(tracking_recorder_class, SQLiteTrackingRecorder)

        recorder = tracking_recorder_class(datastore=self.datastore)

        if self.env_create_table():
            recorder.create_table()
        return recorder

    def process_recorder(self) -> ProcessRecorder:
        process_recorder_topic = self.env.get(self.PROCESS_RECORDER_TOPIC)

        if process_recorder_topic:
            process_recorder_class: type[SQLiteProcessRecorder] = resolve_topic(
                process_recorder_topic
            )
            assert issubclass(process_recorder_class, SQLiteProcessRecorder)
        else:
            process_recorder_class = self.process_recorder_class

        recorder = process_recorder_class(datastore=self.datastore)

        if self.env_create_table():
            recorder.create_table()
        return recorder

    def env_create_table(self) -> bool:
        default = "yes"
        return bool(strtobool(self.env.get(self.CREATE_TABLE, default) or default))

    def close(self) -> None:
        self.datastore.close()


Factory = SQLiteFactory



from __future__ import annotations

import contextlib
from copy import deepcopy
from typing import TYPE_CHECKING

from eventsourcing.dcb.api import (
    DCBAppendCondition,
    DCBEvent,
    DCBQuery,
    DCBReadResponse,
    DCBRecorder,
    DCBSequencedEvent,
)
from eventsourcing.dcb.persistence import (
    DCBInfrastructureFactory,
    DCBListenNotifySubscription,
)
from eventsourcing.persistence import IntegrityError, ProgrammingError
from eventsourcing.popo import POPOFactory, POPORecorder, POPOTrackingRecorder

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from threading import Event


class InMemoryDCBRecorder(DCBRecorder, POPORecorder):
    def __init__(self) -> None:
        super().__init__()
        self.events: list[DCBSequencedEvent] = []
        self.position_sequence = self._position_sequence_generator()
        self._listeners: set[Event] = set()

    def read(
        self,
        query: DCBQuery | None = None,
        *,
        after: int | None = None,
        limit: int | None = None,
    ) -> DCBReadResponse:
        query = query or DCBQuery()
        with self._database_lock:
            events_generator = (
                event
                for event in self.events
                if (after is None or event.position > after)
                and (
                    not query.items
                    or any(
                        (not item.types or event.event.type in item.types)
                        and (set(event.event.tags) >= set(item.tags))
                        for item in query.items
                    )
                )
            )

            events = []
            for i, event in enumerate(events_generator):
                if limit is not None and i >= limit:
                    break
                events.append(deepcopy(event))
            if limit is None:
                head = self.events[-1].position if self.events else None
            else:
                head = events[-1].position if events else None
            # TODO: Change the previous few lines to actually be an iterator.
            return SimpleDCBReadResponse(iter(events), head)

    def append(
        self, events: Sequence[DCBEvent], condition: DCBAppendCondition | None = None
    ) -> int:
        if len(events) == 0:
            msg = "Should be at least one event. Avoid this elsewhere"
            raise ProgrammingError(msg)
        with self._database_lock:
            if condition is not None:
                read_response = self.read(
                    query=condition.fail_if_events_match,
                    after=condition.after,
                    limit=1,
                )
                try:
                    next(read_response)
                except StopIteration:
                    pass
                else:
                    raise IntegrityError(condition)
            self.events.extend(
                DCBSequencedEvent(
                    position=next(self.position_sequence),
                    event=deepcopy(event),
                )
                for event in events
            )
            self._notify_listeners()
            return self.events[-1].position

    def _position_sequence_generator(self) -> Iterator[int]:
        position = 1
        while True:
            yield position
            position += 1

    def subscribe(
        self,
        query: DCBQuery | None = None,
        *,
        after: int | None = None,
    ) -> InMemorySubscription:
        return InMemorySubscription(self, query=query, after=after)

    def listen(self, event: Event) -> None:
        self._listeners.add(event)

    def unlisten(self, event: Event) -> None:
        with contextlib.suppress(KeyError):
            self._listeners.remove(event)

    def _notify_listeners(self) -> None:
        for listener in self._listeners:
            listener.set()


class InMemorySubscription(DCBListenNotifySubscription[InMemoryDCBRecorder]):
    def __init__(
        self,
        recorder: InMemoryDCBRecorder,
        query: DCBQuery | None = None,
        after: int | None = None,
    ) -> None:
        super().__init__(recorder=recorder, query=query, after=after)
        self._recorder.listen(self._has_been_notified)

    def stop(self) -> None:
        super().stop()
        self._recorder.unlisten(self._has_been_notified)


class SimpleDCBReadResponse(DCBReadResponse):
    def __init__(self, events: Iterator[DCBSequencedEvent], head: int | None = None):
        self.events = events
        self._head_was_given = head is not None
        self._head = head

    @property
    def head(self) -> int | None:
        return self._head

    def __next__(self) -> DCBSequencedEvent:
        event = next(self.events)
        if not self._head_was_given:  # pragma: no cover
            self._head = event.position
        return event

    # def next_batch(self) -> list[DCBSequencedEvent]:
    #     """
    #     Returns a batch of events as a list.
    #     Updates the head position similar to __next__.
    #     """
    #     result = []
    #     max_batch_size = 100
    #
    #     # Get up to max_batch_size events from the iterator
    #     try:
    #         for _ in range(max_batch_size):
    #             event = next(self.events)
    #             if not self._head_was_given:
    #                 self._head = event.position
    #             result.append(event)
    #     except StopIteration:
    #         pass
    #     return result


class InMemoryDCBFactory(POPOFactory, DCBInfrastructureFactory[POPOTrackingRecorder]):

    def dcb_recorder(self) -> DCBRecorder:
        return InMemoryDCBRecorder()



from __future__ import annotations

import json
import threading
from unittest import TestCase
from uuid import uuid4

from eventsourcing.dcb.api import (
    DCBAppendCondition,
    DCBEvent,
    DCBQuery,
    DCBQueryItem,
    DCBRecorder,
    DCBSequencedEvent,
    DCBSubscription,
    TDCBRecorder_co,
)
from eventsourcing.persistence import IntegrityError


class DCBRecorderTestCase(TestCase):

    def _test_append_read(
        self, recorder: DCBRecorder, initial_position: int = 0
    ) -> None:
        # Read all, expect no results.
        read_response = recorder.read()
        result = list(read_response)
        self.assertEqual(initial_position, len(list(result)))
        self.assertEqual(initial_position, read_response.head or 0)

        # Append one event.
        event1 = DCBEvent(type="type1", data=b"data1", tags=["tagX"])
        position = recorder.append(events=[event1])

        # Check the returned position is 1.
        self.assertEqual(1 + initial_position, position)

        # Read all, expect one event.
        read_response = recorder.read(after=initial_position)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event1.data, result[0].event.data)
        self.assertEqual(1 + initial_position, read_response.head)

        # Read all after 1, expect no events.
        read_response = recorder.read(after=1 + initial_position)
        result = list(read_response)
        self.assertEqual(0, len(result))
        self.assertEqual(1 + initial_position, read_response.head)

        # Read all limit 1, expect one event.
        read_response = recorder.read(after=initial_position, limit=1)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event1.data, result[0].event.data)
        self.assertEqual(1 + initial_position, read_response.head)

        # Read all limit 0, expect no events (and read_response.head is None).
        read_response = recorder.read(limit=0)
        result = list(read_response)
        self.assertEqual(0, len(result))
        self.assertEqual(None, read_response.head)

        # Read events with type1, expect 1 event.
        query_type1 = DCBQuery(items=[DCBQueryItem(types=["type1"])])
        read_response = recorder.read(query_type1, after=initial_position)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event1.data, result[0].event.data)
        self.assertEqual(1 + initial_position, read_response.head)

        # Read events with type2, expect no events.
        query_type2 = DCBQuery(items=[DCBQueryItem(types=["type2"])])
        read_response = recorder.read(query_type2, after=initial_position)
        result = list(read_response)
        self.assertEqual(0, len(result))
        self.assertEqual(1 + initial_position, read_response.head)

        # Read events with tagX, expect one event.
        query_tag_x = DCBQuery(items=[DCBQueryItem(tags=["tagX"])])
        read_response = recorder.read(query_tag_x, after=initial_position)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event1.data, result[0].event.data)
        self.assertEqual(1 + initial_position, read_response.head)

        # Read events with tagY, expect no events.
        query_tag_y = DCBQuery(items=[DCBQueryItem(tags=["tagY"])])
        read_response = recorder.read(query=query_tag_y, after=initial_position)
        result = list(read_response)
        self.assertEqual(0, len(result), result)
        self.assertEqual(1 + initial_position, read_response.head)

        # Read events with type1 and tagX, expect one event.
        query_type1_tag_x = DCBQuery(
            items=[DCBQueryItem(types=["type1"], tags=["tagX"])]
        )
        read_response = recorder.read(query_type1_tag_x, after=initial_position)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(1 + initial_position, read_response.head)

        # Read events with type1 and tagY, expect no events.
        query_type1_tag_y = DCBQuery(
            items=[DCBQueryItem(types=["type1"], tags=["tagY"])]
        )
        read_response = recorder.read(query_type1_tag_y, after=initial_position)
        result = list(read_response)
        self.assertEqual(0, len(result))
        self.assertEqual(1 + initial_position, read_response.head)

        # Read events with type2 and tagX, expect no events.
        query_type2_tag_x = DCBQuery(
            items=[DCBQueryItem(types=["type2"], tags=["tagX"])]
        )
        read_response = recorder.read(query_type2_tag_x, after=initial_position)
        result = list(read_response)
        self.assertEqual(0, len(result))
        self.assertEqual(1 + initial_position, read_response.head)

        # Append two more events.
        event2 = DCBEvent(type="type2", data=b"data2", tags=["tagA", "tagB"])
        event3 = DCBEvent(type="type3", data=b"data3", tags=["tagA", "tagC"])
        position = recorder.append(events=[event2, event3])

        # Check the returned position is 3
        self.assertEqual(3 + initial_position, position)

        # Read all, expect 3 events (in ascending order).
        read_response = recorder.read(after=initial_position)
        result = list(read_response)
        self.assertEqual(3, len(result))
        self.assertEqual(event1.data, result[0].event.data)
        self.assertEqual(event2.data, result[1].event.data)
        self.assertEqual(event3.data, result[2].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Read all after 1, expect two events.
        read_response = recorder.read(after=1 + initial_position)
        result = list(read_response)
        self.assertEqual(2, len(result))
        self.assertEqual(event2.data, result[0].event.data)
        self.assertEqual(event3.data, result[1].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Read all after 2, expect one event.
        read_response = recorder.read(after=2 + initial_position)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event3.data, result[0].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Read all after 1, limit 1, expect one event.
        read_response = recorder.read(after=1 + initial_position, limit=1)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event2.data, result[0].event.data)
        self.assertEqual(2 + initial_position, read_response.head)

        # Read type1 after 1, expect no events.
        read_response = recorder.read(query_type1, after=1 + initial_position)
        result = list(read_response)
        self.assertEqual(0, len(result))
        self.assertEqual(3 + initial_position, read_response.head)

        # Read tagX after 1, expect no events.
        read_response = recorder.read(query_tag_x, after=1 + initial_position)
        result = list(read_response)
        self.assertEqual(0, len(result))
        self.assertEqual(3 + initial_position, read_response.head)

        # Read type1 and tagX after 1, expect no events.
        read_response = recorder.read(query_type1_tag_x, after=1 + initial_position)
        result = list(read_response)
        self.assertEqual(0, len(result))
        self.assertEqual(3 + initial_position, read_response.head)

        # Read events with tagA, expect two events.
        query_tag_a = DCBQuery(items=[DCBQueryItem(tags=["tagA"])])
        read_response = recorder.read(query_tag_a, after=initial_position)
        result = list(read_response)
        self.assertEqual(2, len(result))
        self.assertEqual(event2.data, result[0].event.data)
        self.assertEqual(event3.data, result[1].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Read events with tagA and tagB, expect one event.
        query_tag_a_and_b = DCBQuery(items=[DCBQueryItem(tags=["tagA", "tagB"])])
        read_response = recorder.read(query_tag_a_and_b, after=initial_position)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event2.data, result[0].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Read events with tagB or tagC, expect two events.
        query_tag_b_or_c = DCBQuery(
            items=[
                DCBQueryItem(tags=["tagB"]),
                DCBQueryItem(tags=["tagC"]),
            ]
        )
        read_response = recorder.read(query_tag_b_or_c, after=initial_position)
        result = list(read_response)
        self.assertEqual(2, len(result))
        self.assertEqual(event2.data, result[0].event.data)
        self.assertEqual(event3.data, result[1].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Read events with tagX or tagY, expect one event.
        query_tag_x_or_y = DCBQuery(
            items=[
                DCBQueryItem(tags=["tagX"]),
                DCBQueryItem(tags=["tagY"]),
            ]
        )
        read_response = recorder.read(query_tag_x_or_y, after=initial_position)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event1.data, result[0].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Read events with type2 and tagA, expect one event.
        query_type2_tag_a = DCBQuery(
            items=[DCBQueryItem(types=["type2"], tags=["tagA"])]
        )
        read_response = recorder.read(query_type2_tag_a, after=initial_position)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event2.data, result[0].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Read events with type2 and tagA after 2, expect no events.
        query_type2_tag_a = DCBQuery(
            items=[DCBQueryItem(types=["type2"], tags=["tagA"])]
        )
        read_response = recorder.read(query_type2_tag_a, after=2 + initial_position)
        result = list(read_response)
        self.assertEqual(0, len(result))
        self.assertEqual(3 + initial_position, read_response.head)

        # Read events with type2 and tagA, expect one event.
        query_type2_tag_a = DCBQuery(
            items=[DCBQueryItem(types=["type2"], tags=["tagA"])]
        )
        read_response = recorder.read(query_type2_tag_a, after=initial_position)
        result = list(read_response)
        self.assertEqual(1, len(result))
        self.assertEqual(event2.data, result[0].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Read events with type2 and tagB, or with type3 and tagC, expect two events.
        query_type2_tag_b_or_type3_tagc = DCBQuery(
            items=[
                DCBQueryItem(types=["type2"], tags=["tagB"]),
                DCBQueryItem(types=["type3"], tags=["tagC"]),
            ]
        )
        read_response = recorder.read(
            query_type2_tag_b_or_type3_tagc, after=initial_position
        )
        result = list(read_response)
        self.assertEqual(2, len(result), result)
        self.assertEqual(event2.data, result[0].event.data)
        self.assertEqual(event3.data, result[1].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Repeat with query items in different order, expect events in ascending order.
        query_type3_tag_c_or_type2_tag_b = DCBQuery(
            items=[
                DCBQueryItem(types=["type3"], tags=["tagC"]),
                DCBQueryItem(types=["type2"], tags=["tagB"]),
            ]
        )
        read_response = recorder.read(
            query_type3_tag_c_or_type2_tag_b, after=initial_position
        )
        result = list(read_response)
        self.assertEqual(2, len(result))
        self.assertEqual(event2.data, result[0].event.data)
        self.assertEqual(event3.data, result[1].event.data)
        self.assertEqual(3 + initial_position, read_response.head)

        # Append must fail if recorded events match condition.
        event4 = DCBEvent(type="type4", data=b"data4")

        # Fail because condition matches all.
        new = [event4]
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition())

        # Fail because condition matches all after 1.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(after=1))

        # Fail because condition matches type1.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_type1))

        # Fail because condition matches type2 after 1.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_type2, after=1))

        # Fail because condition matches tagX.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_tag_x))

        # Fail because condition matches tagA after 1.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_tag_a, after=1))

        # Fail because condition matches type1 and tagX.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_type1_tag_x))

        # Fail because condition matches type2 and tagA after 1.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_type2_tag_a, after=1))

        # Fail because condition matches tagA and tagB.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_tag_a_and_b))

        # Fail because condition matches tagB or tagC.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_tag_b_or_c))

        # Fail because condition matches tagX or tagY.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_tag_x_or_y))

        # Fail because condition matches with type2 and tagB, or with type3 and tagC.
        with self.assertRaises(IntegrityError):
            recorder.append(new, DCBAppendCondition(query_type2_tag_b_or_type3_tagc))

        # Can append after 3.
        recorder.append(new)

        # Can append match type_n.
        query_type_n = DCBQuery(items=[DCBQueryItem(types=["typeN"])])
        recorder.append(new, DCBAppendCondition(query_type_n))

        # Can append match tagY.
        recorder.append(new, DCBAppendCondition(query_tag_y))

        # Can append match type1 after 1.
        recorder.append(
            new, DCBAppendCondition(query_type1, after=1 + initial_position)
        )

        # Can append match tagX after 1.
        recorder.append(
            new, DCBAppendCondition(query_tag_x, after=1 + initial_position)
        )

        # Can append match type1 and tagX after 1.
        recorder.append(
            new, DCBAppendCondition(query_type1_tag_x, after=1 + initial_position)
        )

        # Can append match tagX, after 1.
        recorder.append(
            new, DCBAppendCondition(query_tag_x, after=1 + initial_position)
        )

        # Check it works with course subscription consistency boundaries and events.
        student_id = f"student1-{uuid4()}"
        student_registered = DCBEvent(
            type="StudentRegistered",
            data=json.dumps({"name": "Student1", "max_courses": 10}).encode(),
            tags=[student_id],
        )
        course_id = f"course1-{uuid4()}"
        course_registered = DCBEvent(
            type="CourseRegistered",
            data=json.dumps({"name": "Course1", "places": 10}).encode(),
            tags=[course_id],
        )
        student_joined_course = DCBEvent(
            type="StudentJoinedCourse",
            data=json.dumps(
                {"student_id": student_id, "course_id": course_id}
            ).encode(),
            tags=[course_id, student_id],
        )

        recorder.append(
            events=[student_registered],
            condition=DCBAppendCondition(
                fail_if_events_match=DCBQuery(
                    items=[
                        DCBQueryItem(
                            tags=student_registered.tags, types=["StudentRegistered"]
                        )
                    ],
                ),
                after=3,
            ),
        )
        recorder.append(
            events=[course_registered],
            condition=DCBAppendCondition(
                fail_if_events_match=DCBQuery(
                    items=[DCBQueryItem(tags=course_registered.tags)],
                ),
                after=3,
            ),
        )
        recorder.append(
            events=[student_joined_course],
            condition=DCBAppendCondition(
                fail_if_events_match=DCBQuery(
                    items=[DCBQueryItem(tags=student_joined_course.tags)],
                ),
                after=3,
            ),
        )

        read_response = recorder.read(after=initial_position)
        result = list(read_response)
        self.assertEqual(13, len(result))
        self.assertEqual(result[-3].event.type, student_registered.type)
        self.assertEqual(result[-2].event.type, course_registered.type)
        self.assertEqual(result[-1].event.type, student_joined_course.type)
        self.assertEqual(result[-3].event.data, student_registered.data)
        self.assertEqual(result[-2].event.data, course_registered.data)
        self.assertEqual(result[-1].event.data, student_joined_course.data)
        self.assertEqual(result[-3].event.tags, student_registered.tags)
        self.assertEqual(result[-2].event.tags, course_registered.tags)
        self.assertEqual(result[-1].event.tags, student_joined_course.tags)
        self.assertEqual(13 + initial_position, read_response.head)

        read_response = recorder.read(
            query=DCBQuery(
                items=[DCBQueryItem(tags=student_registered.tags)],
            ),
            after=initial_position,
        )
        self.assertEqual(2, len(list(read_response)))
        self.assertEqual(13 + initial_position, read_response.head)

        read_response = recorder.read(
            query=DCBQuery(
                items=[DCBQueryItem(tags=course_registered.tags)],
            ),
            after=initial_position,
        )
        self.assertEqual(2, len(list(read_response)))
        self.assertEqual(13 + initial_position, read_response.head)

        read_response = recorder.read(
            query=DCBQuery(
                items=[DCBQueryItem(tags=student_joined_course.tags)],
            ),
            after=initial_position,
        )
        self.assertEqual(1, len(list(read_response)))
        self.assertEqual(13 + initial_position, read_response.head)

        read_response = recorder.read(
            query=DCBQuery(
                items=[DCBQueryItem(tags=student_registered.tags)],
            ),
            after=2 + initial_position,
        )
        self.assertEqual(2, len(list(read_response)))
        self.assertEqual(13 + initial_position, read_response.head)

        read_response = recorder.read(
            query=DCBQuery(
                items=[DCBQueryItem(tags=course_registered.tags)],
            ),
            after=2 + initial_position,
        )
        self.assertEqual(2, len(list(read_response)))
        self.assertEqual(13 + initial_position, read_response.head)

        read_response = recorder.read(
            query=DCBQuery(
                items=[DCBQueryItem(tags=student_joined_course.tags)],
            ),
            after=2 + initial_position,
        )
        self.assertEqual(1, len(list(read_response)))
        self.assertEqual(13 + initial_position, read_response.head)

        read_response = recorder.read(
            query=DCBQuery(
                items=[DCBQueryItem(tags=student_registered.tags)],
            ),
            after=2 + initial_position,
            limit=1,
        )
        self.assertEqual(1, len(list(read_response)))
        self.assertEqual(11 + initial_position, read_response.head)

        read_response = recorder.read(
            query=DCBQuery(
                items=[DCBQueryItem(tags=course_registered.tags)],
            ),
            after=2 + initial_position,
            limit=1,
        )
        self.assertEqual(1, len(list(read_response)))
        self.assertEqual(12 + initial_position, read_response.head)

        read_response = recorder.read(
            query=DCBQuery(
                items=[DCBQueryItem(tags=student_joined_course.tags)],
            ),
            after=2 + initial_position,
            limit=1,
        )
        self.assertEqual(1, len(list(read_response)))
        self.assertEqual(13 + initial_position, read_response.head)

        consistency_boundary = DCBQuery(
            items=[
                DCBQueryItem(
                    types=["StudentRegistered", "StudentJoinedCourse"],
                    tags=[student_id],
                ),
                DCBQueryItem(
                    types=["CourseRegistered", "StudentJoinedCourse"],
                    tags=[course_id],
                ),
            ]
        )
        read_response = recorder.read(
            query=consistency_boundary,
            after=initial_position,
        )
        self.assertEqual(3, len(list(read_response)))
        self.assertEqual(13 + initial_position, read_response.head)

        # # Check it works with appointment booking.
        # appointment_id = f"appointment-{uuid4()}"
        # tags = [
        #     f"slot:2025-07-10-{hour:02}-{minute:02}"
        #     for hour in range(13, 18)
        #     for minute in range(0, 60)
        # ]
        # appointment_scheduled = DCBEvent(
        #     type="AppointmentSchedules",
        #     data=json.dumps({"name": "ABC"}).encode(),
        #     tags=tags,
        # )
        # started = datetime.datetime.now()
        # eventstore.append(
        #     events=[appointment_scheduled],
        #     condition=DCBAppendCondition(
        #         fail_if_events_match=DCBQuery(
        #             items=[DCBQueryItem(tags=tags)],
        #         ),
        #     ),
        # )
        # print("Event appended:", datetime.datetime.now() - started)
        # started = datetime.datetime.now()
        # with self.assertRaises(IntegrityError):
        #     eventstore.append(
        #         events=[appointment_scheduled],
        #         condition=DCBAppendCondition(
        #             fail_if_events_match=DCBQuery(
        #                 items=[DCBQueryItem(tags=tags)],
        #             ),
        #         ),
        #     )
        # print("Conflict detected:", datetime.datetime.now() - started)

    def _test_append_subscribe(
        self, recorder: DCBRecorder, initial_position: int = 0
    ) -> None:
        # Append one event.
        event1 = DCBEvent(type="type1", data=b"data1", tags=["tagX"])
        position1 = recorder.append(events=[event1])
        self.assertEqual(1 + initial_position, position1)

        # Start subscription.
        with recorder.subscribe(after=initial_position) as subscription:
            self.assertEqual(position1, next(subscription).position)

            thread = EnsureSubscriptionBlockAndReceive(subscription)
            thread.has_blocked.wait()
            self.assertFalse(thread.has_received.wait(timeout=0.5))

            # Append one more event.
            event2 = DCBEvent(type="type1", data=b"data1", tags=["tagX"])
            position2 = recorder.append(events=[event2])
            self.assertEqual(2 + initial_position, position2)

            self.assertTrue(thread.has_received.wait(timeout=1))
            assert thread.received_event is not None  # for mypy
            self.assertEqual(position2, thread.received_event.position)


class EnsureSubscriptionBlockAndReceive(threading.Thread):
    def __init__(self, subscription: DCBSubscription[TDCBRecorder_co]):
        super().__init__()
        self.subscription = subscription
        self.has_blocked = threading.Event()
        self.has_received = threading.Event()
        self.received_event: DCBSequencedEvent | None = None
        self.start()

    def run(self) -> None:
        self.has_blocked.set()
        self.received_event = next(self.subscription)
        self.has_received.set()



from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal
from typing import cast
from uuid import uuid4

from eventsourcing.domain import Aggregate, AggregateCreated, AggregateEvent


@dataclass(frozen=True)
class EmailAddress:
    address: str


class BankAccount(Aggregate):
    """Aggregate root for bank accounts."""

    def __init__(self, full_name: str, email_address: EmailAddress):
        self.full_name = full_name
        self.email_address = email_address
        self.balance = Decimal("0.00")
        self.overdraft_limit = Decimal("0.00")
        self.is_closed = False

    @classmethod
    def open(cls, full_name: str, email_address: str) -> BankAccount:
        """Creates new bank account object."""
        return cls._create(
            cls.Opened,
            id=uuid4(),
            full_name=full_name,
            email_address=EmailAddress(email_address),
        )

    class Opened(AggregateCreated):
        full_name: str
        email_address: str

    def append_transaction(self, amount: Decimal) -> None:
        """Appends given amount as transaction on account."""
        self.check_account_is_not_closed()
        self.check_has_sufficient_funds(amount)
        self.trigger_event(
            self.TransactionAppended,
            amount=amount,
        )

    def check_account_is_not_closed(self) -> None:
        if self.is_closed:
            raise AccountClosedError({"account_id": self.id})

    def check_has_sufficient_funds(self, amount: Decimal) -> None:
        if self.balance + amount < -self.overdraft_limit:
            raise InsufficientFundsError({"account_id": self.id})

    @dataclass(frozen=True)
    class TransactionAppended(AggregateEvent):
        """Domain event for when transaction
        is appended to bank account.
        """

        amount: Decimal

        def apply(self, aggregate: Aggregate) -> None:
            """Increments the account balance."""
            cast("BankAccount", aggregate).balance += self.amount

    def set_overdraft_limit(self, overdraft_limit: Decimal) -> None:
        """Sets the overdraft limit."""
        # Check the limit is not a negative value.
        assert overdraft_limit >= Decimal("0.00")
        self.check_account_is_not_closed()
        self.trigger_event(
            self.OverdraftLimitSet,
            overdraft_limit=overdraft_limit,
        )

    class OverdraftLimitSet(AggregateEvent):
        """Domain event for when overdraft
        limit is set.
        """

        overdraft_limit: Decimal

        def apply(self, aggregate: Aggregate) -> None:
            cast("BankAccount", aggregate).overdraft_limit = self.overdraft_limit

    def close(self) -> None:
        """Closes the bank account."""
        self.trigger_event(self.Closed)

    class Closed(AggregateEvent):
        """Domain event for when account is closed."""

        def apply(self, aggregate: Aggregate) -> None:
            cast("BankAccount", aggregate).is_closed = True


class AccountClosedError(Exception):
    """Raised when attempting to operate a closed account."""


class InsufficientFundsError(Exception):
    """Raised when attempting to go past overdraft limit."""
